import logging
from typing import Any, Optional

from mlflow.data.dataset_source import DatasetSource
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_managed_catalog_messages_pb2 import (
    GetTable,
    GetTableResponse,
)
from mlflow.protos.databricks_managed_catalog_service_pb2 import DatabricksUnityCatalogService
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.utils._spark_utils import _get_active_spark_session
from mlflow.utils._unity_catalog_utils import get_full_name_from_sc
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow.utils.proto_json_utils import message_to_json
from mlflow.utils.rest_utils import (
    _REST_API_PATH_PREFIX,
    call_endpoint,
    extract_api_info_for_service,
)
from mlflow.utils.string_utils import _backtick_quote

DATABRICKS_HIVE_METASTORE_NAME = "hive_metastore"
# these two catalog names both points to the workspace local default HMS (hive metastore).
DATABRICKS_LOCAL_METASTORE_NAMES = [DATABRICKS_HIVE_METASTORE_NAME, "spark_catalog"]
# samples catalog is managed by databricks for hosting public dataset like NYC taxi dataset.
# it is neither a UC nor local metastore catalog
DATABRICKS_SAMPLES_CATALOG_NAME = "samples"

_logger = logging.getLogger(__name__)


class DeltaDatasetSource(DatasetSource):
    """
    Represents the source of a dataset stored at in a delta table.
    """

    def __init__(
        self,
        path: Optional[str] = None,
        delta_table_name: Optional[str] = None,
        delta_table_version: Optional[int] = None,
        delta_table_id: Optional[str] = None,
    ):
        if (path, delta_table_name).count(None) != 1:
            raise MlflowException(
                'Must specify exactly one of "path" or "table_name"',
                INVALID_PARAMETER_VALUE,
            )
        self._path = path
        if delta_table_name is not None:
            self._delta_table_name = get_full_name_from_sc(
                delta_table_name, _get_active_spark_session()
            )
        else:
            self._delta_table_name = delta_table_name
        self._delta_table_version = delta_table_version
        self._delta_table_id = delta_table_id

    @staticmethod
    def _get_source_type() -> str:
        return "delta_table"

    def load(self, **kwargs):
        """
        Loads the dataset source as a Delta Dataset Source.

        Returns:
            An instance of ``pyspark.sql.DataFrame``.
        """
        from pyspark.sql import SparkSession

        spark = SparkSession.builder.getOrCreate()

        spark_read_op = spark.read.format("delta")
        if self._delta_table_version is not None:
            spark_read_op = spark_read_op.option("versionAsOf", self._delta_table_version)

        if self._path:
            return spark_read_op.load(self._path)
        else:
            backticked_delta_table_name = ".".join(
                map(_backtick_quote, self._delta_table_name.split("."))
            )
            return spark_read_op.table(backticked_delta_table_name)

    @property
    def path(self) -> Optional[str]:
        return self._path

    @property
    def delta_table_name(self) -> Optional[str]:
        return self._delta_table_name

    @property
    def delta_table_id(self) -> Optional[str]:
        return self._delta_table_id

    @property
    def delta_table_version(self) -> Optional[int]:
        return self._delta_table_version

    @staticmethod
    def _can_resolve(raw_source: Any):
        return False

    @classmethod
    def _resolve(cls, raw_source: str) -> "DeltaDatasetSource":
        raise NotImplementedError

    # check if table is in the Databricks Unity Catalog
    def _is_databricks_uc_table(self):
        if self._delta_table_name is not None:
            catalog_name = self._delta_table_name.split(".", 1)[0]
            return (
                catalog_name not in DATABRICKS_LOCAL_METASTORE_NAMES
                and catalog_name != DATABRICKS_SAMPLES_CATALOG_NAME
            )
        else:
            return False

    def _lookup_table_id(self, table_name):
        try:
            req_body = message_to_json(GetTable(full_name_arg=table_name))
            _METHOD_TO_INFO = extract_api_info_for_service(
                DatabricksUnityCatalogService, _REST_API_PATH_PREFIX
            )
            db_creds = get_databricks_host_creds()
            endpoint, method = _METHOD_TO_INFO[GetTable]
            # We need to replace the full_name_arg in the endpoint definition with
            # the actual table name for the REST API to work.
            final_endpoint = endpoint.replace("{full_name_arg}", table_name)
            resp = call_endpoint(
                host_creds=db_creds,
                endpoint=final_endpoint,
                method=method,
                json_body=req_body,
                response_proto=GetTableResponse,
            )
            return resp.table_id
        except Exception:
            return None

    def to_dict(self) -> dict[Any, Any]:
        info = {}
        if self._path:
            info["path"] = self._path
        if self._delta_table_name:
            info["delta_table_name"] = self._delta_table_name
        if self._delta_table_version:
            info["delta_table_version"] = self._delta_table_version
        if self._is_databricks_uc_table():
            info["is_databricks_uc_table"] = True
            if self._delta_table_id:
                info["delta_table_id"] = self._delta_table_id
            else:
                info["delta_table_id"] = self._lookup_table_id(self._delta_table_name)
        return info

    @classmethod
    def from_dict(cls, source_dict: dict[Any, Any]) -> "DeltaDatasetSource":
        return cls(
            path=source_dict.get("path"),
            delta_table_name=source_dict.get("delta_table_name"),
            delta_table_version=source_dict.get("delta_table_version"),
            delta_table_id=source_dict.get("delta_table_id"),
        )



from dataclasses import dataclass
from typing import Optional

from mlflow.genai.scorers.base import Scorer
from mlflow.utils.annotations import experimental

_ERROR_MSG = (
    "The `databricks-agents` package is required to use `mlflow.genai.scheduled_scorers`. "
    "Please install it with `pip install databricks-agents`."
)


@experimental(version="3.0.0")
@dataclass()
class ScorerScheduleConfig:
    """
    A scheduled scorer configuration for automated monitoring of generative AI applications.

    Scheduled scorers are used to automatically evaluate traces logged to MLflow experiments
    by production applications. They are part of [Databricks Lakehouse Monitoring for GenAI](https://docs.databricks.com/aws/en/generative-ai/agent-evaluation/monitoring),
    which helps track quality metrics like groundedness, safety, and guideline adherence
    alongside operational metrics like volume, latency, and cost.

    When configured, scheduled scorers run automatically in the background to evaluate
    a sample of traces based on the specified sampling rate and filter criteria. The
    Assessments are displayed in the Traces tab of the MLflow experiment and can be used to
    identify quality issues in production.

    Args:
        scorer: The scorer function to run on sampled traces. Must be either a built-in
            scorer (e.g., Safety, Correctness) or a function decorated with @scorer.
            Subclasses of Scorer are not supported.
        scheduled_scorer_name: The name for this scheduled scorer configuration
            within the experiment. This name must be unique among all scheduled scorers
            in the same experiment.
            We recommend using the scorer's name (e.g., scorer.name) for consistency.
        sample_rate: The fraction of traces to evaluate, between 0.0 and 1.0. For example,
            0.1 means 10% of traces will be randomly selected for evaluation.
        filter_string: An optional MLflow search_traces compatible filter string to apply
            before sampling traces. Only traces matching this filter will be considered
            for evaluation. Uses the same syntax as mlflow.search_traces().

    Example:
        .. code-block:: python

            from mlflow.genai.scorers import Safety, scorer
            from mlflow.genai.scheduled_scorers import ScorerScheduleConfig

            # Using a built-in scorer
            safety_config = ScorerScheduleConfig(
                scorer=Safety(),
                scheduled_scorer_name="production_safety",
                sample_rate=0.2,  # Evaluate 20% of traces
                filter_string="trace.status = 'OK'",
            )


            # Using a custom scorer
            @scorer
            def response_length(outputs):
                return len(str(outputs)) > 100


            length_config = ScorerScheduleConfig(
                scorer=response_length,
                scheduled_scorer_name="adequate_length",
                sample_rate=0.1,  # Evaluate 10% of traces
                filter_string="trace.status = 'OK'",
            )

    Note:
        Scheduled scorers are executed automatically by Databricks and do not need to be
        manually triggered. The Assessments appear in the Traces tab of the MLflow
        experiment. Only traces logged directly to the experiment are monitored; traces
        logged to individual runs within the experiment are not evaluated.

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """

    scorer: Scorer
    scheduled_scorer_name: str
    sample_rate: float
    filter_string: Optional[str] = None


# Scheduled Scorer CRUD operations
@experimental(version="3.0.0")
def add_scheduled_scorer(  # clint: disable=missing-docstring-param  # noqa: D417
    *,
    scheduled_scorer_name: str,
    scorer: Scorer,
    sample_rate: float,
    filter_string: Optional[str] = None,
    experiment_id: Optional[str] = None,
    **kwargs,
) -> ScorerScheduleConfig:
    """
    Add a scheduled scorer to automatically monitor traces in an MLflow experiment.

    This function configures a scorer function to run automatically on traces logged to the
    specified experiment. The scorer will evaluate a sample of traces based on the sampling rate
    and any filter criteria. Assessments are displayed in the Traces tab of the MLflow experiment.

    Args:
        scheduled_scorer_name: The name for this scheduled scorer within the experiment.
            We recommend using the scorer's name (e.g., scorer.name) for consistency.
        scorer: The scorer function to execute on sampled traces. Must be either a
            built-in scorer or a function decorated with @scorer. Subclasses of Scorer
            are not supported.
        sample_rate: The fraction of traces to evaluate, between 0.0 and 1.0. For example,
            0.3 means 30% of traces will be randomly selected for evaluation.
        filter_string: An optional MLflow search_traces compatible filter string. Only
            traces matching this filter will be considered for evaluation. If None,
            all traces in the experiment are eligible for sampling.
        experiment_id: The ID of the MLflow experiment to monitor. If None, uses the
            currently active experiment.

    Returns:
        A ScorerScheduleConfig object representing the configured scheduled scorer.

    Example:
        .. code-block:: python

            import mlflow
            from mlflow.genai.scorers import Safety, Correctness
            from mlflow.genai.scheduled_scorers import add_scheduled_scorer

            # Set up your experiment
            experiment = mlflow.set_experiment("my_genai_app_monitoring")

            # Add a safety scorer to monitor 50% of traces
            safety_scorer = add_scheduled_scorer(
                scheduled_scorer_name="safety_monitor",
                scorer=Safety(),
                sample_rate=0.5,
                filter_string="trace.status = 'OK'",
            )

            # Add a correctness scorer with different sampling
            correctness_scorer = add_scheduled_scorer(
                scheduled_scorer_name="correctness_monitor",
                scorer=Correctness(),
                sample_rate=0.2,  # More expensive, so lower sample rate
                experiment_id=experiment.experiment_id,  # Explicitly specify experiment
            )

    Note:
        Once added, the scheduled scorer will begin evaluating new traces automatically.
        There may be a delay between when traces are logged and when they are evaluated.
        Only traces logged directly to the experiment are monitored; traces logged to
        individual runs within the experiment are not evaluated.

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import add_scheduled_scorer
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return add_scheduled_scorer(
        experiment_id, scheduled_scorer_name, scorer, sample_rate, filter_string, **kwargs
    )


@experimental(version="3.0.0")
def update_scheduled_scorer(  # clint: disable=missing-docstring-param  # noqa: D417
    *,
    scheduled_scorer_name: str,
    scorer: Optional[Scorer] = None,
    sample_rate: Optional[float] = None,
    filter_string: Optional[str] = None,
    experiment_id: Optional[str] = None,
    **kwargs,
) -> ScorerScheduleConfig:
    """
    Update an existing scheduled scorer configuration.

    This function modifies the configuration of an existing scheduled scorer, allowing you
    to change the scorer function, sampling rate, or filter criteria. Only the provided
    parameters will be updated; omitted parameters will retain their current values.
    The scorer will continue to run automatically with the new configuration.

    Args:
        scheduled_scorer_name: The name of the existing scheduled scorer to update. Must match
            the name used when the scorer was originally added. We recommend using the
            scorer's name (e.g., scorer.name) for consistency.
        scorer: The new scorer function to execute on sampled traces. Must be either
            a built-in scorer or a function decorated with @scorer. If None, the
            current scorer function will be retained.
        sample_rate: The new fraction of traces to evaluate, between 0.0 and 1.0.
            If None, the current sample rate will be retained.
        filter_string: The new MLflow search_traces compatible filter string. If None,
            the current filter string will be retained. Pass an empty string to remove
            the filter entirely.
        experiment_id: The ID of the MLflow experiment containing the scheduled scorer.
            If None, uses the currently active experiment.

    Returns:
        A ScorerScheduleConfig object representing the updated scheduled scorer configuration.

    Example:
        .. code-block:: python

            from mlflow.genai.scorers import Safety
            from mlflow.genai.scheduled_scorers import update_scheduled_scorer

            # Update an existing safety scorer to increase sampling rate
            updated_scorer = update_scheduled_scorer(
                scheduled_scorer_name="safety_monitor",
                sample_rate=0.8,  # Increased from 0.5 to 0.8
            )

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import update_scheduled_scorer
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return update_scheduled_scorer(
        experiment_id, scheduled_scorer_name, scorer, sample_rate, filter_string, **kwargs
    )


@experimental(version="3.0.0")
def delete_scheduled_scorer(  # clint: disable=missing-docstring-param  # noqa: D417
    *,
    scheduled_scorer_name: str,
    experiment_id: Optional[str] = None,
    **kwargs,
) -> None:
    """
    Delete a scheduled scorer from an MLflow experiment.

    This function removes a scheduled scorer configuration, stopping automatic evaluation
    of traces. Existing Assessments will remain in the Traces tab of the MLflow
    experiment, but no new evaluations will be performed.

    Args:
        scheduled_scorer_name: The name of the scheduled scorer to delete. Must match the name
            used when the scorer was originally added.
        experiment_id: The ID of the MLflow experiment containing the scheduled scorer.
            If None, uses the currently active experiment.

    Example:
        .. code-block:: python

            from mlflow.genai.scheduled_scorers import delete_scheduled_scorer

            # Remove a scheduled scorer that's no longer needed
            delete_scheduled_scorer(scheduled_scorer_name="safety_monitor")

            # To delete all scheduled scorers at once, use set_scheduled_scorers
            # with an empty list instead:
            from mlflow.genai.scheduled_scorers import set_scheduled_scorers

            set_scheduled_scorers(
                scheduled_scorers=[]  # Empty list removes all scorers
            )

    Note:
        Deletion is immediate and cannot be undone. If you need the same scorer
        configuration later, you will need to add it again using add_scheduled_scorer.

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import delete_scheduled_scorer
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return delete_scheduled_scorer(experiment_id, scheduled_scorer_name, **kwargs)


@experimental(version="3.0.0")
def get_scheduled_scorer(  # clint: disable=missing-docstring-param  # noqa: D417
    *,
    scheduled_scorer_name: str,
    experiment_id: Optional[str] = None,
    **kwargs,
) -> ScorerScheduleConfig:
    """
    Retrieve the configuration of a specific scheduled scorer.

    This function returns the current configuration of a scheduled scorer, including
    its scorer function, sampling rate, and filter criteria.

    Args:
        scheduled_scorer_name: The name of the scheduled scorer to retrieve.
        experiment_id: The ID of the MLflow experiment containing the scheduled scorer.
            If None, uses the currently active experiment.

    Returns:
        A ScorerScheduleConfig object containing the current configuration of the specified
        scheduled scorer.

    Example:
        .. code-block:: python

            from mlflow.genai.scheduled_scorers import get_scheduled_scorer

            # Get the current configuration of a scheduled scorer
            scorer_config = get_scheduled_scorer(scheduled_scorer_name="safety_monitor")

            print(f"Sample rate: {scorer_config.sample_rate}")
            print(f"Filter: {scorer_config.filter_string}")
            print(f"Scorer: {scorer_config.scorer.name}")

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import get_scheduled_scorer
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return get_scheduled_scorer(experiment_id, scheduled_scorer_name, **kwargs)


@experimental(version="3.0.0")
def list_scheduled_scorers(  # clint: disable=missing-docstring-param  # noqa: D417
    *, experiment_id: Optional[str] = None, **kwargs
) -> list[ScorerScheduleConfig]:
    """
    List all scheduled scorers for an experiment.

    This function returns all scheduled scorers configured for the specified experiment,
    or for the current active experiment if no experiment ID is provided.

    Args:
        experiment_id: The ID of the MLflow experiment to list scheduled scorers for.
            If None, uses the currently active experiment.

    Returns:
        A list of ScheduledScorerConfig objects representing all scheduled scorers configured
        for the specified experiment.

    Example:
        .. code-block:: python

            import mlflow
            from mlflow.genai.scheduled_scorers import list_scheduled_scorers

            # List scorers for a specific experiment
            scorers = list_scheduled_scorers(experiment_id="12345")
            for scorer in scorers:
                print(f"Scorer: {scorer.scheduled_scorer_name}")
                print(f"Sample rate: {scorer.sample_rate}")
                print(f"Filter: {scorer.filter_string}")

            # List scorers for the current active experiment
            mlflow.set_experiment("my_genai_app_monitoring")
            current_scorers = list_scheduled_scorers()
            print(f"Found {len(current_scorers)} scheduled scorers")

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import list_scheduled_scorers
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return list_scheduled_scorers(experiment_id, **kwargs)


@experimental(version="3.0.0")
def set_scheduled_scorers(  # clint: disable=missing-docstring-param  # noqa: D417
    *,
    scheduled_scorers: list[ScorerScheduleConfig],
    experiment_id: Optional[str] = None,
    **kwargs,
) -> None:
    """
    Replace all scheduled scorers for an experiment with the provided list.

    This function removes all existing scheduled scorers for the specified experiment
    and replaces them with the new list. This is useful for batch configuration updates
    or when you want to ensure only specific scorers are active.

    Args:
        scheduled_scorers: A list of ScheduledScorerConfig objects to set as the complete
            set of scheduled scorers for the experiment. Any existing scheduled scorers
            not in this list will be removed.
        experiment_id: The ID of the MLflow experiment to configure. If None, uses the
            currently active experiment.

    Example:
        .. code-block:: python

            from mlflow.genai.scorers import Safety, Correctness, RelevanceToQuery
            from mlflow.genai.scheduled_scorers import ScorerScheduleConfig, set_scheduled_scorers

            # Define a complete monitoring configuration
            monitoring_config = [
                ScorerScheduleConfig(
                    scorer=Safety(),
                    scheduled_scorer_name="safety_check",
                    sample_rate=1.0,  # Check all traces for safety
                ),
                ScorerScheduleConfig(
                    scorer=Correctness(),
                    scheduled_scorer_name="correctness_check",
                    sample_rate=0.2,  # Sample 20% for correctness (more expensive)
                    filter_string="trace.status = 'OK'",
                ),
                ScorerScheduleConfig(
                    scorer=RelevanceToQuery(),
                    scheduled_scorer_name="relevance_check",
                    sample_rate=0.5,  # Sample 50% for relevance
                ),
            ]

            # Apply this configuration, replacing any existing scorers
            set_scheduled_scorers(scheduled_scorers=monitoring_config)

    Warning:
        This function will remove all existing scheduled scorers for the experiment
        that are not included in the provided list. Use add_scheduled_scorer() if you
        want to add scorers without affecting existing ones.

    Note:
        Existing Assessments will remain in the Traces tab of the MLflow experiment.

    .. warning::
        This API is in Beta and may change or be removed in a future release without warning.
    """
    try:
        from databricks.agents.scorers import set_scheduled_scorers
    except ImportError as e:
        raise ImportError(_ERROR_MSG) from e
    return set_scheduled_scorers(experiment_id, scheduled_scorers, **kwargs)



"""
Genesis-Flow Secure Model Loading

This module provides secure alternatives to pickle-based model loading,
addressing security vulnerabilities in model deserialization.
"""

import io
import pickle
import cloudpickle
import logging
import hashlib
from typing import Any, Set, Optional, Union, Type
from pathlib import Path

logger = logging.getLogger(__name__)

# Allowlist of safe classes that can be unpickled
SAFE_PICKLE_CLASSES = {
    # NumPy types
    'numpy.ndarray',
    'numpy.dtype',
    'numpy.int32', 'numpy.int64', 'numpy.float32', 'numpy.float64',
    'numpy.bool_', 'numpy.str_',

    # Pandas types
    'pandas.core.frame.DataFrame',
    'pandas.core.series.Series',
    'pandas.core.index.Index',
    'pandas.core.dtypes.dtypes.CategoricalDtype',

    # Scikit-learn estimators
    'sklearn.linear_model._base.LinearRegression',
    'sklearn.linear_model._logistic.LogisticRegression',
    'sklearn.ensemble._forest.RandomForestClassifier',
    'sklearn.ensemble._forest.RandomForestRegressor',
    'sklearn.tree._classes.DecisionTreeClassifier',
    'sklearn.tree._classes.DecisionTreeRegressor',
    'sklearn.svm._classes.SVC',
    'sklearn.svm._classes.SVR',

    # Built-in types
    'builtins.dict', 'builtins.list', 'builtins.tuple', 'builtins.set',
    'builtins.str', 'builtins.int', 'builtins.float', 'builtins.bool',
    'builtins.type',

    # Collections
    'collections.OrderedDict',
    'collections.defaultdict',

    # MLflow types
    'mlflow.models.signature.ModelSignature',
    'mlflow.models.signature._TypeHints',
    'mlflow.types.schema.Schema',
    'mlflow.pyfunc.model.PythonModel',

    # Cloudpickle internals
    'cloudpickle.cloudpickle._make_skeleton_class',
    'cloudpickle.cloudpickle._class_setstate',
    'cloudpickle.cloudpickle._make_function',
    'cloudpickle.cloudpickle._builtin_type',
    'cloudpickle.cloudpickle._function_setstate',
    'cloudpickle.cloudpickle._make_empty_cell',
    'cloudpickle.cloudpickle._make_cell',

    # Sentence Transformers
    'sentence_transformers.SentenceTransformer.SentenceTransformer',
    'sentence_transformers.model_card.SentenceTransformerModelCardData',
    'sentence_transformers.models.Transformer.Transformer',
    'sentence_transformers.models.Pooling.Pooling',
    'sentence_transformers.models.Normalize.Normalize',

    # Torch
    'torch.torch_version.TorchVersion',
    'torch._utils._rebuild_tensor_v2',
    'torch.storage._load_from_bytes',
    'torch.nn.modules.sparse.Embedding',
    'torch._utils._rebuild_parameter',
    'torch.nn.modules.normalization.LayerNorm',
    'torch.nn.modules.dropout.Dropout',
    'torch.nn.modules.container.ModuleList',
    'torch.nn.modules.linear.Linear',
    'torch.nn.modules.activation.Tanh',
    'torch.float32',
    'torch._C._nn.gelu',

    # Transformers
    'transformers.models.bert.modeling_bert.BertModel',
    'transformers.models.bert.modeling_bert.BertEmbeddings',
    'transformers.models.bert.modeling_bert.BertEncoder',
    'transformers.models.bert.modeling_bert.BertLayer',
    'transformers.models.bert.modeling_bert.BertAttention',
    'transformers.models.bert.modeling_bert.BertSdpaSelfAttention',
    'transformers.models.bert.modeling_bert.BertSelfOutput',
    'transformers.models.bert.modeling_bert.BertIntermediate',
    'transformers.models.bert.modeling_bert.BertOutput',
    'transformers.models.bert.modeling_bert.BertPooler',
    'transformers.models.bert.configuration_bert.BertConfig',
    'transformers.models.bert.tokenization_bert_fast.BertTokenizerFast',
    'transformers.activations.GELUActivation',

    # Tokenizers
    'tokenizers.Tokenizer',
    'tokenizers.models.Model',
    'tokenizers.AddedToken',
}


class RestrictedUnpickler(pickle.Unpickler):
    """
    Secure unpickler that only allows safe, whitelisted classes.
    
    This prevents arbitrary code execution during model deserialization
    by restricting which classes can be instantiated.
    """
    
    def __init__(self, file, *, safe_classes: Optional[Set[str]] = None):
        super().__init__(file)
        self.safe_classes = safe_classes or SAFE_PICKLE_CLASSES
        
    def find_class(self, module: str, name: str) -> Type:
        """
        Override to restrict class loading to safe classes only.
        
        Args:
            module: Module name
            name: Class name
            
        Returns:
            Class object if safe
            
        Raises:
            SecurityError: If class is not in allowlist
        """
        full_name = f"{module}.{name}"
        
        # Check if the class is in our safe list
        if full_name in self.safe_classes:
            logger.debug(f"Loading safe class: {full_name}")
            return super().find_class(module, name)
        
        # Additional check for known safe modules
        safe_modules = {
            'numpy': ['ndarray', 'dtype', 'int32', 'int64', 'float32', 'float64', 'bool_'],
            'pandas': ['DataFrame', 'Series', 'Index'],
            'builtins': ['dict', 'list', 'tuple', 'set', 'str', 'int', 'float', 'bool'],
        }
        
        if module in safe_modules and name in safe_modules[module]:
            logger.debug(f"Loading safe module class: {full_name}")
            return super().find_class(module, name)
        
        # Log and block unsafe class
        logger.warning(f"Blocked potentially unsafe class: {full_name}")
        raise pickle.UnpicklingError(
            f"Security: Class '{full_name}' is not in the allowlist. "
            f"If this is a legitimate model class, add it to SAFE_PICKLE_CLASSES."
        )

class SecureModelLoader:
    """
    Secure model loading with multiple safety mechanisms.
    """
    
    @staticmethod
    def calculate_file_hash(file_path: Union[str, Path]) -> str:
        """Calculate SHA256 hash of a file for integrity checking."""
        hasher = hashlib.sha256()
        with open(file_path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hasher.update(chunk)
        return hasher.hexdigest()
    
    @staticmethod
    def safe_pickle_load(file_path: Union[str, Path], *, safe_classes: Optional[Set[str]] = None) -> Any:
        """
        Safely load a pickle file using restricted unpickler.
        
        Args:
            file_path: Path to pickle file
            safe_classes: Optional custom set of safe classes
            
        Returns:
            Unpickled object
            
        Raises:
            SecurityError: If file contains unsafe classes
            FileNotFoundError: If file doesn't exist
        """
        file_path = Path(file_path)
        
        if not file_path.exists():
            raise FileNotFoundError(f"Model file not found: {file_path}")
        
        # Log file hash for audit trail
        file_hash = SecureModelLoader.calculate_file_hash(file_path)
        logger.info(f"Loading model file {file_path.name} (SHA256: {file_hash[:16]}...)")
        
        try:
            with open(file_path, 'rb') as f:
                unpickler = RestrictedUnpickler(f, safe_classes=safe_classes)
                model = unpickler.load()
                logger.info(f"Successfully loaded model of type: {type(model).__name__}")
                return model
        except pickle.UnpicklingError as e:
            logger.error(f"Security: Unsafe model file rejected: {e}")
            raise SecurityError(f"Model file contains unsafe content: {e}") from e
        except Exception as e:
            logger.error(f"Error loading model: {e}")
            raise
    
    @staticmethod
    def safe_cloudpickle_load(file_path: Union[str, Path], *, safe_classes: Optional[Set[str]] = None) -> Any:
        """
        Safely load a cloudpickle file using restricted unpickler.
        
        Args:
            file_path: Path to cloudpickle file
            safe_classes: Optional custom set of safe classes
            
        Returns:
            Unpickled object
        """
        file_path = Path(file_path)
        
        if not file_path.exists():
            raise FileNotFoundError(f"Model file not found: {file_path}")
        
        # Calculate file hash for audit trail
        file_hash = SecureModelLoader.calculate_file_hash(file_path)
        logger.info(f"Loading cloudpickle file {file_path.name} (SHA256: {file_hash[:16]}...)")
        
        try:
            with open(file_path, 'rb') as f:
                # Use cloudpickle's load with our custom unpickler
                # Note: This is a simplified approach - in practice, cloudpickle's
                # load function would need to be modified to accept a custom unpickler
                unpickler = RestrictedUnpickler(f, safe_classes=safe_classes)
                model = unpickler.load()
                logger.info(f"Successfully loaded cloudpickle model of type: {type(model).__name__}")
                return model
        except pickle.UnpicklingError as e:
            logger.error(f"Security: Unsafe cloudpickle file rejected: {e}")
            raise SecurityError(f"CloudPickle file contains unsafe content: {e}") from e
        except Exception as e:
            logger.error(f"Error loading cloudpickle model: {e}")
            raise

class SecurityError(Exception):
    """Exception raised for security-related model loading issues."""
    pass

def add_safe_class(class_name: str) -> None:
    """
    Add a class to the safe loading allowlist.
    
    Args:
        class_name: Full class name (e.g., 'mymodule.MyClass')
    """
    SAFE_PICKLE_CLASSES.add(class_name)
    logger.info(f"Added {class_name} to safe loading allowlist")

def remove_safe_class(class_name: str) -> None:
    """
    Remove a class from the safe loading allowlist.
    
    Args:
        class_name: Full class name to remove
    """
    SAFE_PICKLE_CLASSES.discard(class_name)
    logger.info(f"Removed {class_name} from safe loading allowlist")

def get_safe_classes() -> Set[str]:
    """Return a copy of the current safe classes set."""
    return SAFE_PICKLE_CLASSES.copy()


# This file contains utility functions for scorer functionality.

import ast
import inspect
import logging
import re
from textwrap import dedent
from typing import Any, Callable, Optional

_logger = logging.getLogger(__name__)


# FunctionBodyExtractor class is forked from https://github.com/unitycatalog/unitycatalog/blob/20dd3820be332ac04deec4e063099fb863eb3392/ai/core/src/unitycatalog/ai/core/utils/callable_utils.py
class FunctionBodyExtractor(ast.NodeVisitor):
    """
    AST NodeVisitor class to extract the body of a function.
    """

    def __init__(self, func_name: str, source_code: str):
        self.func_name = func_name
        self.source_code = source_code
        self.function_body = ""
        self.indent_unit = 4
        self.found = False

    def visit_FunctionDef(self, node: ast.FunctionDef):
        if not self.found and node.name == self.func_name:
            self.found = True
            self.extract_body(node)

    def extract_body(self, node: ast.FunctionDef):
        body = node.body
        # Skip the docstring
        if (
            body
            and isinstance(body[0], ast.Expr)
            and isinstance(body[0].value, ast.Constant)
            and isinstance(body[0].value.value, str)
        ):
            body = body[1:]

        if not body:
            return

        start_lineno = body[0].lineno
        end_lineno = body[-1].end_lineno

        source_lines = self.source_code.splitlines(keepends=True)
        function_body_lines = source_lines[start_lineno - 1 : end_lineno]

        self.function_body = dedent("".join(function_body_lines)).rstrip("\n")

        indents = [stmt.col_offset for stmt in body if stmt.col_offset is not None]
        if indents:
            self.indent_unit = min(indents)


# extract_function_body function is forked from https://github.com/unitycatalog/unitycatalog/blob/20dd3820be332ac04deec4e063099fb863eb3392/ai/core/src/unitycatalog/ai/core/utils/callable_utils.py
def extract_function_body(func: Callable[..., Any]) -> tuple[str, int]:
    """
    Extracts the body of a function as a string without the signature or docstring,
    dedents the code, and returns the indentation unit used in the function (e.g., 2 or 4 spaces).
    """
    source_lines, _ = inspect.getsourcelines(func)
    dedented_source = dedent("".join(source_lines))
    func_name = func.__name__

    extractor = FunctionBodyExtractor(func_name, dedented_source)
    parsed_source = ast.parse(dedented_source)
    extractor.visit(parsed_source)

    return extractor.function_body, extractor.indent_unit


def recreate_function(source: str, signature: str, func_name: str) -> Optional[Callable[..., Any]]:
    """
    Recreate a function from its source code, signature, and name.

    Args:
        source: The function body source code.
        signature: The function signature string (e.g., "(inputs, outputs)").
        func_name: The name of the function.

    Returns:
        The recreated function or None if recreation failed.
    """
    try:
        # Parse the signature to build the function definition
        sig_match = re.match(r"\((.*?)\)", signature)
        if not sig_match:
            return None

        params_str = sig_match.group(1).strip()

        # Build the function definition
        func_def = f"def {func_name}({params_str}):\n"
        # Indent the source code
        indented_source = "\n".join(f"    {line}" for line in source.split("\n"))
        func_def += indented_source

        # Create a namespace with common MLflow imports that scorer functions might use
        import_namespace = {}

        # Import commonly used MLflow classes
        try:
            from mlflow.entities import (
                Assessment,
                AssessmentError,
                AssessmentSource,
                AssessmentSourceType,
                Feedback,
            )
            from mlflow.genai.judges import CategoricalRating

            import_namespace.update(
                {
                    "Feedback": Feedback,
                    "Assessment": Assessment,
                    "AssessmentSource": AssessmentSource,
                    "AssessmentError": AssessmentError,
                    "AssessmentSourceType": AssessmentSourceType,
                    "CategoricalRating": CategoricalRating,
                }
            )
        except ImportError:
            pass  # Some imports might not be available in all contexts

        local_namespace = {}

        # Execute the function definition with MLflow imports available
        exec(func_def, import_namespace, local_namespace)

        # Return the recreated function
        return local_namespace[func_name]

    except Exception:
        _logger.warning(f"Failed to recreate function '{func_name}' from serialized source code")
        return None



import logging
import warnings
from typing import Optional

from mlflow.tracking.context.abstract_context import RunContextProvider
from mlflow.tracking.context.databricks_cluster_context import DatabricksClusterRunContext
from mlflow.tracking.context.databricks_command_context import DatabricksCommandRunContext
from mlflow.tracking.context.databricks_job_context import DatabricksJobRunContext
from mlflow.tracking.context.databricks_notebook_context import DatabricksNotebookRunContext
from mlflow.tracking.context.databricks_repo_context import DatabricksRepoRunContext
from mlflow.tracking.context.default_context import DefaultRunContext
from mlflow.tracking.context.git_context import GitRunContext
from mlflow.tracking.context.system_environment_context import SystemEnvironmentContext
from mlflow.utils.plugins import get_entry_points

_logger = logging.getLogger(__name__)


class RunContextProviderRegistry:
    """Registry for run context provider implementations

    This class allows the registration of a run context provider which can be used to infer meta
    information about the context of an MLflow experiment run. Implementations declared though the
    entrypoints `mlflow.run_context_provider` group can be automatically registered through the
    `register_entrypoints` method.

    Registered run context providers can return tags that override those implemented in the core
    library, however the order in which plugins are resolved is undefined.
    """

    def __init__(self):
        self._registry = []

    def register(self, run_context_provider_cls):
        self._registry.append(run_context_provider_cls())

    def register_entrypoints(self):
        """Register tracking stores provided by other packages"""
        for entrypoint in get_entry_points("mlflow.run_context_provider"):
            try:
                self.register(entrypoint.load())
            except (AttributeError, ImportError) as exc:
                warnings.warn(
                    'Failure attempting to register context provider "{}": {}'.format(
                        entrypoint.name, str(exc)
                    ),
                    stacklevel=2,
                )

    def __iter__(self):
        return iter(self._registry)


_run_context_provider_registry = RunContextProviderRegistry()
_run_context_provider_registry.register(DefaultRunContext)
_run_context_provider_registry.register(GitRunContext)
_run_context_provider_registry.register(DatabricksNotebookRunContext)
_run_context_provider_registry.register(DatabricksJobRunContext)
_run_context_provider_registry.register(DatabricksClusterRunContext)
_run_context_provider_registry.register(DatabricksCommandRunContext)
_run_context_provider_registry.register(DatabricksRepoRunContext)
_run_context_provider_registry.register(SystemEnvironmentContext)

_run_context_provider_registry.register_entrypoints()


def resolve_tags(tags=None, ignore: Optional[list[RunContextProvider]] = None):
    """Generate a set of tags for the current run context. Tags are resolved in the order,
    contexts are registered. Argument tags are applied last.

    This function iterates through all run context providers in the registry. Additional context
    providers can be registered as described in
    :py:class:`mlflow.tracking.context.RunContextProvider`.

    Args:
        tags: A dictionary of tags to override. If specified, tags passed in this argument will
            override those inferred from the context.
        ignore: A list of RunContextProvider classes to exclude from the resolution.

    Returns:
        A dictionary of resolved tags.
    """
    ignore = ignore or []
    all_tags = {}
    for provider in _run_context_provider_registry:
        if any(isinstance(provider, ig) for ig in ignore):
            continue

        try:
            if provider.in_context():
                all_tags.update(provider.tags())
        except Exception as e:
            _logger.warning("Encountered unexpected error during resolving tags: %s", e)

    if tags is not None:
        all_tags.update(tags)

    return all_tags



"""
Databricks Agent Labeling Python SDK. For more details see Databricks Agent Evaluation:
<https://docs.databricks.com/en/generative-ai/agent-evaluation/index.html>

The API docs can be found here:
<https://api-docs.databricks.com/python/databricks-agents/latest/databricks_agent_eval.html#review-app>
"""

from typing import Any, Optional

from mlflow.genai.labeling.labeling import Agent, LabelingSession, ReviewApp

_ERROR_MSG = (
    "The `databricks-agents` package is required to use `mlflow.genai.labeling`. "
    "Please install it with `pip install databricks-agents`."
)


def get_review_app(experiment_id: Optional[str] = None) -> "ReviewApp":
    """Gets or creates (if it doesn't exist) the review app for the given experiment ID.

    .. note::
        This functionality is only available in Databricks. Please run
        `pip install mlflow[databricks]` to use it.

    Args:
        experiment_id: Optional. The experiment ID for which to get the review app. If not provided,
            the experiment ID is inferred from the current active environment.

    Returns:
        ReviewApp: The review app.
    """
    try:
        from databricks.agents.review_app import get_review_app as _get_review_app
    except ImportError:
        raise ImportError(_ERROR_MSG) from None

    return ReviewApp(_get_review_app(experiment_id))


def create_labeling_session(
    name: str,
    *,
    assigned_users: Optional[list[str]] = None,
    agent: Optional[str] = None,
    label_schemas: Optional[list[str]] = None,
    enable_multi_turn_chat: bool = False,
    custom_inputs: Optional[dict[str, Any]] = None,
) -> LabelingSession:
    """Create a new labeling session in the review app.

    .. note::
        This functionality is only available in Databricks. Please run
        `pip install mlflow[databricks]` to use it.

    Args:
        name: The name of the labeling session.
        assigned_users: The users that will be assigned to label items in the session.
        agent: The agent to be used to generate responses for the items in the session.
        label_schemas: The label schemas to be used in the session.
        enable_multi_turn_chat: Whether to enable multi-turn chat labeling for the session.
        custom_inputs: Optional. Custom inputs to be used in the session.

    Returns:
        LabelingSession: The created labeling session.
    """
    try:
        from databricks.agents.review_app import get_review_app as _get_review_app
    except ImportError:
        raise ImportError(_ERROR_MSG) from None

    return LabelingSession(
        _get_review_app().create_labeling_session(
            name=name,
            assigned_users=assigned_users or [],
            agent=agent,
            label_schemas=label_schemas or [],
            enable_multi_turn_chat=enable_multi_turn_chat,
            custom_inputs=custom_inputs,
        )
    )


def get_labeling_sessions() -> list[LabelingSession]:
    """Get all labeling sessions from the review app.

    .. note::
        This functionality is only available in Databricks. Please run
        `pip install mlflow[databricks]` to use it.

    Returns:
        list[LabelingSession]: The list of labeling sessions.
    """
    try:
        from databricks.agents.review_app import get_review_app as _get_review_app
    except ImportError:
        raise ImportError(_ERROR_MSG) from None

    return [LabelingSession(session) for session in _get_review_app().get_labeling_sessions()]


def get_labeling_session(run_id: str) -> LabelingSession:
    """Get a labeling session from the review app.

    .. note::
        This functionality is only available in Databricks. Please run
        `pip install mlflow[databricks]` to use it.

    Args:
        run_id: The mlflow run ID of the labeling session to get.

    Returns:
        LabelingSession: The labeling session.
    """
    labeling_sessions = get_labeling_sessions()
    labeling_session = next(
        (
            labeling_session
            for labeling_session in labeling_sessions
            if labeling_session.mlflow_run_id == run_id
        ),
        None,
    )
    if labeling_session is None:
        raise ValueError(f"Labeling session with run_id `{run_id}` not found")
    return LabelingSession(labeling_session)


def delete_labeling_session(labeling_session: LabelingSession) -> "ReviewApp":
    """Delete a labeling session from the review app.

    .. note::
        This functionality is only available in Databricks. Please run
        `pip install mlflow[databricks]` to use it.

    Args:
        labeling_session: The labeling session to delete.

    Returns:
        ReviewApp: The review app.
    """
    try:
        from databricks.agents.review_app import get_review_app as _get_review_app
    except ImportError:
        raise ImportError(_ERROR_MSG) from None

    return ReviewApp(_get_review_app().delete_labeling_session(labeling_session._session))


__all__ = [
    "Agent",
    "LabelingSession",
    "ReviewApp",
    "get_review_app",
    "create_labeling_session",
    "get_labeling_sessions",
    "get_labeling_session",
    "delete_labeling_session",
]



"""
Scoring server for python model format.
The passed int model is expected to have function:
   predict(pandas.Dataframe) -> pandas.DataFrame

Input, expected in text/csv or application/json format,
is parsed into pandas.DataFrame and passed to the model.

Defines four endpoints:
    /ping used for health check
    /health (same as /ping)
    /version used for getting the mlflow version
    /invocations used for scoring
"""

import asyncio
import inspect
import json
import logging
import os
import shlex
import sys
import traceback
from functools import wraps
from typing import Any, NamedTuple, Optional

from mlflow.environment_variables import (
    _MLFLOW_IS_IN_SERVING_ENVIRONMENT,
    MLFLOW_SCORING_SERVER_REQUEST_TIMEOUT,
)

# NB: We need to be careful what we import form mlflow here. Scoring server is used from within
# model's conda environment. The version of mlflow doing the serving (outside) and the version of
# mlflow in the model's conda environment (inside) can differ. We should therefore keep mlflow
# dependencies to the minimum here.
# ALl of the mlflow dependencies below need to be backwards compatible.
from mlflow.exceptions import MlflowException
from mlflow.pyfunc.model import _log_warning_if_params_not_in_predict_signature
from mlflow.types import ParamSchema, Schema
from mlflow.utils import reraise
from mlflow.utils.file_utils import path_to_local_file_uri
from mlflow.utils.proto_json_utils import (
    MlflowInvalidInputException,
    NumpyEncoder,
    _get_jsonable_obj,
    dataframe_from_parsed_json,
    parse_tf_serving_input,
)
from mlflow.version import VERSION

try:
    from mlflow.pyfunc import PyFuncModel, load_model
except ImportError:
    from mlflow.pyfunc import load_pyfunc as load_model
from io import StringIO

from mlflow.protos.databricks_pb2 import BAD_REQUEST, INVALID_PARAMETER_VALUE
from mlflow.pyfunc.utils.serving_data_parser import is_unified_llm_input

_SERVER_MODEL_PATH = "__pyfunc_model_path__"
SERVING_MODEL_CONFIG = "SERVING_MODEL_CONFIG"

CONTENT_TYPE_CSV = "text/csv"
CONTENT_TYPE_JSON = "application/json"

CONTENT_TYPES = [
    CONTENT_TYPE_CSV,
    CONTENT_TYPE_JSON,
]

_logger = logging.getLogger(__name__)

DF_RECORDS = "dataframe_records"
DF_SPLIT = "dataframe_split"
INSTANCES = "instances"
INPUTS = "inputs"

SUPPORTED_FORMATS = {DF_RECORDS, DF_SPLIT, INSTANCES, INPUTS}
SERVING_PARAMS_KEY = "params"

REQUIRED_INPUT_FORMAT = (
    f"The input must be a JSON dictionary with exactly one of the input fields {SUPPORTED_FORMATS}"
)
SCORING_PROTOCOL_CHANGE_INFO = (
    "IMPORTANT: The MLflow Model scoring protocol has changed in MLflow version 2.0. If you are"
    " seeing this error, you are likely using an outdated scoring request format. To resolve the"
    " error, either update your request format or adjust your MLflow Model's requirements file to"
    " specify an older version of MLflow (for example, change the 'mlflow' requirement specifier"
    " to 'mlflow==1.30.0'). If you are making a request using the MLflow client"
    " (e.g. via `mlflow.pyfunc.spark_udf()`), upgrade your MLflow client to a version >= 2.0 in"
    " order to use the new request format. For more information about the updated MLflow"
    " Model scoring protocol in MLflow 2.0, see"
    " https://mlflow.org/docs/latest/models.html#deploy-mlflow-models."
)


def load_model_with_mlflow_config(model_uri):
    extra_kwargs = {}
    if model_config_json := os.environ.get(SERVING_MODEL_CONFIG):
        extra_kwargs["model_config"] = json.loads(model_config_json)

    return load_model(model_uri, **extra_kwargs)


# Keep this method to maintain compatibility with MLServer
# https://github.com/SeldonIO/MLServer/blob/caa173ab099a4ec002a7c252cbcc511646c261a6/runtimes/mlflow/mlserver_mlflow/runtime.py#L13C5-L13C31
def infer_and_parse_json_input(json_input, schema: Schema = None):
    """
    Args:
        json_input: A JSON-formatted string representation of TF serving input or a Pandas
                    DataFrame, or a stream containing such a string representation.
        schema: Optional schema specification to be used during parsing.
    """
    if isinstance(json_input, dict):
        decoded_input = json_input
    else:
        try:
            decoded_input = json.loads(json_input)
        except json.decoder.JSONDecodeError as ex:
            raise MlflowException(
                message=(
                    "Failed to parse input from JSON. Ensure that input is a valid JSON"
                    f" formatted string. Error: '{ex}'. Input: \n{json_input}\n"
                ),
                error_code=BAD_REQUEST,
            )
    if isinstance(decoded_input, dict):
        format_keys = set(decoded_input.keys()).intersection(SUPPORTED_FORMATS)
        if len(format_keys) != 1:
            message = f"Received dictionary with input fields: {list(decoded_input.keys())}"
            raise MlflowException(
                message=f"{REQUIRED_INPUT_FORMAT}. {message}. {SCORING_PROTOCOL_CHANGE_INFO}",
                error_code=BAD_REQUEST,
            )
        input_format = format_keys.pop()
        if input_format in (INSTANCES, INPUTS):
            return parse_tf_serving_input(decoded_input, schema=schema)

        elif input_format in (DF_SPLIT, DF_RECORDS):
            # NB: skip the dataframe_ prefix
            pandas_orient = input_format[10:]
            return dataframe_from_parsed_json(
                decoded_input[input_format], pandas_orient=pandas_orient, schema=schema
            )
    elif isinstance(decoded_input, list):
        message = "Received a list"
        raise MlflowException(
            message=f"{REQUIRED_INPUT_FORMAT}. {message}. {SCORING_PROTOCOL_CHANGE_INFO}",
            error_code=BAD_REQUEST,
        )
    else:
        message = f"Received unexpected input type '{type(decoded_input)}'"
        raise MlflowException(
            message=f"{REQUIRED_INPUT_FORMAT}. {message}.", error_code=BAD_REQUEST
        )


def _decode_json_input(json_input):
    """
    Args:
        json_input: A JSON-formatted string representation of TF serving input or a Pandas
                    DataFrame, or a stream containing such a string representation.

    Returns:
        A dictionary representation of the JSON input.
    """
    if isinstance(json_input, dict):
        return json_input

    try:
        decoded_input = json.loads(json_input)
    except json.decoder.JSONDecodeError as ex:
        raise MlflowInvalidInputException(
            "Ensure that input is a valid JSON formatted string. "
            f"Error: '{ex!r}'\nInput: \n{json_input}\n"
        ) from ex

    if isinstance(decoded_input, dict):
        return decoded_input
    if isinstance(decoded_input, list):
        raise MlflowInvalidInputException(f"{REQUIRED_INPUT_FORMAT}. Received a list.")

    raise MlflowInvalidInputException(
        f"{REQUIRED_INPUT_FORMAT}. Received unexpected input type '{type(decoded_input)}."
    )


def _split_data_and_params_for_llm_input(json_input, param_schema: Optional[ParamSchema]):
    data = {}
    params = {}
    schema_params = {param.name for param in param_schema.params} if param_schema else {}

    for key, value in json_input.items():
        # if the model defines a param schema, then we can add
        # it to the params dict. otherwise, add it to the data
        # dict to prevent it from being ignored at inference time
        if key in schema_params:
            params[key] = value
        else:
            data[key] = value

    return data, params


def _split_data_and_params(json_input):
    input_dict = _decode_json_input(json_input)
    data = {k: v for k, v in input_dict.items() if k in SUPPORTED_FORMATS}
    params = input_dict.pop(SERVING_PARAMS_KEY, None)
    return data, params


def infer_and_parse_data(data, schema: Schema = None):
    """
    Args:
        data: A dictionary representation of TF serving input or a Pandas
            DataFrame, or a stream containing such a string representation.
        schema: Optional schema specification to be used during parsing.
    """

    format_keys = set(data.keys()).intersection(SUPPORTED_FORMATS)
    if len(format_keys) != 1:
        message = f"Received dictionary with input fields: {list(data.keys())}"
        raise MlflowException(
            message=f"{REQUIRED_INPUT_FORMAT}. {message}. {SCORING_PROTOCOL_CHANGE_INFO}",
            error_code=BAD_REQUEST,
        )
    input_format = format_keys.pop()
    if input_format in (INSTANCES, INPUTS):
        return parse_tf_serving_input(data, schema=schema)

    if input_format in (DF_SPLIT, DF_RECORDS):
        pandas_orient = input_format[10:]  # skip the dataframe_ prefix
        return dataframe_from_parsed_json(
            data[input_format], pandas_orient=pandas_orient, schema=schema
        )


def parse_csv_input(csv_input, schema: Schema = None):
    """
    Args:
        csv_input: A CSV-formatted string representation of a Pandas DataFrame, or a stream
                   containing such a string representation.
        schema: Optional schema specification to be used during parsing.
    """
    import pandas as pd

    try:
        if schema is None:
            return pd.read_csv(csv_input)
        else:
            dtypes = dict(zip(schema.input_names(), schema.pandas_types()))
            return pd.read_csv(csv_input, dtype=dtypes)
    except Exception as e:
        _handle_serving_error(
            error_message=(
                "Failed to parse input as a Pandas DataFrame. Ensure that the input is"
                " a valid CSV-formatted Pandas DataFrame produced using the"
                f" `pandas.DataFrame.to_csv()` method. Error: '{e}'"
            ),
            error_code=BAD_REQUEST,
        )


def unwrapped_predictions_to_json(raw_predictions, output):
    predictions = _get_jsonable_obj(raw_predictions, pandas_orient="records")
    return json.dump(predictions, output, cls=NumpyEncoder)


def predictions_to_json(raw_predictions, output, metadata=None):
    if metadata and "predictions" in metadata:
        raise MlflowException(
            "metadata cannot contain 'predictions' key", error_code=INVALID_PARAMETER_VALUE
        )
    predictions = _get_jsonable_obj(raw_predictions, pandas_orient="records")
    return json.dump({"predictions": predictions, **(metadata or {})}, output, cls=NumpyEncoder)


def _handle_serving_error(error_message, error_code, include_traceback=True):
    """
    Logs information about an exception thrown by model inference code that is currently being
    handled and reraises it with the specified error message. The exception stack trace
    is also included in the reraised error message.

    Args:
        error_message: A message for the reraised exception.
        error_code: An appropriate error code for the reraised exception. This should be one of
            the codes listed in the `mlflow.protos.databricks_pb2` proto.
        include_traceback: Whether to include the current traceback in the returned error.
    """
    if include_traceback:
        traceback_buf = StringIO()
        traceback.print_exc(file=traceback_buf)
        traceback_str = traceback_buf.getvalue()
        e = MlflowException(message=error_message, error_code=error_code, stack_trace=traceback_str)
    else:
        e = MlflowException(message=error_message, error_code=error_code)
    reraise(MlflowException, e)


class InvocationsResponse(NamedTuple):
    response: str
    status: int
    mimetype: str


def invocations(data, content_type, model, input_schema):
    type_parts = list(map(str.strip, content_type.split(";")))
    mime_type = type_parts[0]
    parameter_value_pairs = type_parts[1:]
    parameter_values = {
        key: value for pair in parameter_value_pairs for key, _, value in [pair.partition("=")]
    }

    charset = parameter_values.get("charset", "utf-8").lower()
    if charset != "utf-8":
        return InvocationsResponse(
            response="The scoring server only supports UTF-8",
            status=415,
            mimetype="text/plain",
        )

    unexpected_content_parameters = set(parameter_values.keys()).difference({"charset"})
    if unexpected_content_parameters:
        return InvocationsResponse(
            response=(
                f"Unrecognized content type parameters: "
                f"{', '.join(unexpected_content_parameters)}. "
                f"{SCORING_PROTOCOL_CHANGE_INFO}"
            ),
            status=415,
            mimetype="text/plain",
        )

    # The traditional JSON request/response format, wraps the data with one of the supported keys
    # like "dataframe_split" and "predictions". For LLM use cases, we also support unwrapped JSON
    # payload, to provide unified prediction interface.
    should_parse_as_unified_llm_input = False

    if mime_type == CONTENT_TYPE_CSV:
        # Convert from CSV to pandas
        if isinstance(data, bytes):
            data = data.decode("utf-8")
        csv_input = StringIO(data)
        data = parse_csv_input(csv_input=csv_input, schema=input_schema)
        params = None
    elif mime_type == CONTENT_TYPE_JSON:
        parsed_json_input = _parse_json_data(data, model.metadata, input_schema)
        data = parsed_json_input.data
        params = parsed_json_input.params
        should_parse_as_unified_llm_input = parsed_json_input.is_unified_llm_input
    else:
        return InvocationsResponse(
            response=(
                "This predictor only supports the following content types:"
                f" Types: {CONTENT_TYPES}."
                f" Got '{content_type}'."
            ),
            status=415,
            mimetype="text/plain",
        )

    # Do the prediction
    # NB: utils._validate_serving_input mimic the scoring process here to validate input_example
    # work for serving, so any changes here should be reflected there as well
    try:
        if "params" in inspect.signature(model.predict).parameters:
            raw_predictions = model.predict(data, params=params)
        else:
            _log_warning_if_params_not_in_predict_signature(_logger, params)
            raw_predictions = model.predict(data)
    except MlflowException as e:
        if "Failed to enforce schema" in e.message:
            _logger.warning(
                "If using `instances` as input key, we internally convert "
                "the data type from `records` (List[Dict]) type to "
                "`list` (Dict[str, List]) type if the data is a pandas "
                "dataframe representation. This might cause schema changes. "
                "Please use `inputs` to avoid this conversion.\n"
            )
        e.message = f"Failed to predict data '{data}'. \nError: {e.message}"
        raise e
    except Exception:
        raise MlflowException(
            message=(
                "Encountered an unexpected error while evaluating the model. Verify"
                " that the serialized input Dataframe is compatible with the model for"
                " inference."
            ),
            error_code=BAD_REQUEST,
            stack_trace=traceback.format_exc(),
        )
    result = StringIO()

    # if the data was formatted using the unified LLM format,
    # then return the data without the "predictions" key
    if should_parse_as_unified_llm_input:
        unwrapped_predictions_to_json(raw_predictions, result)
    else:
        predictions_to_json(raw_predictions, result)

    return InvocationsResponse(response=result.getvalue(), status=200, mimetype="application/json")


class ParsedJsonInput(NamedTuple):
    data: Any
    params: Optional[dict[str, Any]]
    is_unified_llm_input: bool


def _parse_json_data(data, metadata, input_schema):
    json_input = _decode_json_input(data)
    _is_unified_llm_input = is_unified_llm_input(json_input)
    # no data parsing for unified LLM input format
    if _is_unified_llm_input:
        # Unified LLM input format
        if hasattr(metadata, "get_params_schema"):
            params_schema = metadata.get_params_schema()
        else:
            params_schema = None
        data, params = _split_data_and_params_for_llm_input(json_input, params_schema)
    else:
        # Traditional json input format
        data, params = _split_data_and_params(data)
        # data only needs to be parsed if the model signature is not from type hint
        # default to True for backwards compatibility
        should_parse_data = (
            not metadata._is_signature_from_type_hint()
            if hasattr(metadata, "_is_signature_from_type_hint")
            else True
        )
        if should_parse_data:
            data = infer_and_parse_data(data, input_schema)
        else:
            if INPUTS not in data:
                raise MlflowException.invalid_parameter_value(
                    "Request payload must be a dictionary with 'inputs' key when "
                    f"the model contains a valid type hint. Found keys in payload: {data.keys()}."
                )
            data = data[INPUTS]
    return ParsedJsonInput(data, params, _is_unified_llm_input)


def _async_catch_mlflow_exception(func):
    from fastapi.responses import Response

    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except MlflowException as e:
            return Response(
                content=e.serialize_as_json(),
                status_code=e.get_http_status_code(),
                media_type="application/json",
            )

    return wrapper


def init(model: PyFuncModel):
    """
    Initialize the server. Loads pyfunc model from the path.
    """
    from fastapi import FastAPI, Request
    from fastapi.responses import Response

    app = FastAPI()
    input_schema = model.metadata.get_input_schema()
    # set the environment variable to indicate that we are in a serving environment
    os.environ[_MLFLOW_IS_IN_SERVING_ENVIRONMENT.name] = "true"
    timeout = MLFLOW_SCORING_SERVER_REQUEST_TIMEOUT.get()

    @app.middleware("http")
    async def timeout_middleware(request: Request, call_next):
        try:
            return await asyncio.wait_for(call_next(request), timeout=timeout)
        except (asyncio.TimeoutError, TimeoutError):
            return Response(
                content="Request processing time exceeded limit",
                status_code=504,
                media_type="application/json",
            )

    @app.route("/ping", methods=["GET"])
    @app.route("/health", methods=["GET"])
    async def ping(request: Request):
        """
        Determine if the container is working and healthy.
        We declare it healthy if we can load the model successfully.
        """
        health = model is not None
        status = 200 if health else 404
        return Response(content="\n", status_code=status, media_type="application/json")

    @app.route("/version", methods=["GET"])
    async def version(request: Request):
        """
        Returns the current mlflow version.
        """
        return Response(content=VERSION, status_code=200, media_type="application/json")

    @app.route("/invocations", methods=["POST"])
    @_async_catch_mlflow_exception
    async def transformation(request: Request):
        """
        Do an inference on a single batch of data. In this sample server,
        we take data as CSV or json, convert it to a Pandas DataFrame or Numpy,
        generate predictions and convert them back to json.
        """

        data = await request.body()
        content_type = request.headers.get("content-type")
        # TODO: convert "invocations" to an async method to make internal logic fully non-blocking.
        result = await asyncio.to_thread(invocations, data, content_type, model, input_schema)

        return Response(
            content=result.response, status_code=result.status, media_type=result.mimetype
        )

    return app


def _predict(model_uri, input_path, output_path, content_type):
    from mlflow.pyfunc.utils.environment import _simulate_serving_environment

    with _simulate_serving_environment():
        pyfunc_model = load_model(model_uri)

        should_parse_as_unified_llm_input = False
        if content_type == "json":
            if input_path is None:
                input_str = sys.stdin.read()
            else:
                with open(input_path) as f:
                    input_str = f.read()
            parsed_json_input = _parse_json_data(
                data=input_str,
                metadata=pyfunc_model.metadata,
                input_schema=pyfunc_model.metadata.get_input_schema(),
            )
            df = parsed_json_input.data
            params = parsed_json_input.params
            should_parse_as_unified_llm_input = parsed_json_input.is_unified_llm_input
        elif content_type == "csv":
            df = (
                parse_csv_input(input_path)
                if input_path is not None
                else parse_csv_input(sys.stdin)
            )
            params = None
        else:
            raise Exception(f"Unknown content type '{content_type}'")

        if "params" in inspect.signature(pyfunc_model.predict).parameters:
            raw_predictions = pyfunc_model.predict(df, params=params)
        else:
            _log_warning_if_params_not_in_predict_signature(_logger, params)
            raw_predictions = pyfunc_model.predict(df)

        parse_output_func = (
            unwrapped_predictions_to_json
            if should_parse_as_unified_llm_input
            else predictions_to_json
        )

        if output_path is None:
            parse_output_func(raw_predictions, sys.stdout)
        else:
            with open(output_path, "w") as fout:
                parse_output_func(raw_predictions, fout)


def _serve(model_uri, port, host):
    pyfunc_model = load_model(model_uri)
    init(pyfunc_model).run(port=port, host=host)


def get_cmd(
    model_uri: str,
    port: Optional[int] = None,
    host: Optional[int] = None,
    timeout: Optional[int] = None,
    nworkers: Optional[int] = None,
) -> tuple[str, dict[str, str]]:
    local_uri = path_to_local_file_uri(model_uri)
    timeout = timeout or MLFLOW_SCORING_SERVER_REQUEST_TIMEOUT.get()

    args = []
    if host:
        args.append(f"--host {shlex.quote(host)}")

    if port:
        args.append(f"--port {port}")

    if nworkers:
        args.append(f"--workers {nworkers}")

    command = f"uvicorn {' '.join(args)} mlflow.pyfunc.scoring_server.app:app"

    command_env = os.environ.copy()
    command_env[_SERVER_MODEL_PATH] = local_uri
    command_env[MLFLOW_SCORING_SERVER_REQUEST_TIMEOUT.name] = str(timeout)

    return command, command_env



from typing import Optional

import mlflow.utils.databricks_utils
from mlflow.environment_variables import MLFLOW_ENABLE_UC_VOLUME_FUSE_ARTIFACT_REPO
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_PARAMETER_VALUE
from mlflow.store.artifact.databricks_sdk_artifact_repo import DatabricksSdkArtifactRepository
from mlflow.store.artifact.local_artifact_repo import LocalArtifactRepository
from mlflow.utils.uri import (
    get_databricks_profile_uri_from_artifact_uri,
    is_databricks_model_registry_artifacts_uri,
    is_valid_uc_volumes_uri,
    remove_databricks_profile_info_from_artifact_uri,
    strip_scheme,
)


class UCVolumesArtifactRepository(DatabricksSdkArtifactRepository):
    """
    Stores artifacts on UC Volumes using the Files REST API.
    """

    def __init__(self, artifact_uri: str, tracking_uri: Optional[str] = None) -> None:
        if not is_valid_uc_volumes_uri(artifact_uri):
            raise MlflowException(
                message=(
                    f"UC volumes URI must be of the form "
                    f"dbfs:/Volumes/<catalog>/<schema>/<volume>/<path>: {artifact_uri}"
                ),
                error_code=INVALID_PARAMETER_VALUE,
            )
        uri = remove_databricks_profile_info_from_artifact_uri(artifact_uri)
        super().__init__("/" + strip_scheme(uri).strip("/"), tracking_uri)


def uc_volume_artifact_repo_factory(artifact_uri: str, tracking_uri: Optional[str] = None):
    """
    Returns an ArtifactRepository subclass for storing artifacts on Volumes.

    This factory method is used with URIs of the form ``dbfs:/Volumes/<path>``. Volume-backed
    artifact storage can only be used together with the RestStore.

    Args:
        artifact_uri: Volume root artifact URI.
        tracking_uri: The tracking URI.

    Returns:
        Subclass of ArtifactRepository capable of storing artifacts on DBFS.
    """
    if not is_valid_uc_volumes_uri(artifact_uri):
        raise MlflowException(
            message=(
                f"UC volumes URI must be of the form "
                f"dbfs:/Volumes/<catalog>/<schema>/<volume>/<path>: {artifact_uri}"
            ),
            error_code=INVALID_PARAMETER_VALUE,
        )

    artifact_uri = artifact_uri.rstrip("/")
    db_profile_uri = get_databricks_profile_uri_from_artifact_uri(artifact_uri)
    if (
        mlflow.utils.databricks_utils.is_uc_volume_fuse_available()
        and MLFLOW_ENABLE_UC_VOLUME_FUSE_ARTIFACT_REPO.get()
        and not is_databricks_model_registry_artifacts_uri(artifact_uri)
        and (db_profile_uri is None or db_profile_uri == "databricks")
    ):
        # If the UC Volume FUSE mount is available, write artifacts directly to
        # /Volumes/... using local filesystem APIs.
        # Note: it is possible for a named Databricks profile to point to the current workspace,
        # but we're going to avoid doing a complex check and assume users will use `databricks`
        # to mean the current workspace. Using `UCVolumesArtifactRepository` to access
        # the current workspace's Volumes should still work; it just may be slower.
        uri_without_profile = remove_databricks_profile_info_from_artifact_uri(artifact_uri)
        path = strip_scheme(uri_without_profile).lstrip("/")
        return LocalArtifactRepository(f"file:///{path}", tracking_uri)
    return UCVolumesArtifactRepository(artifact_uri, tracking_uri)



import shutil
import subprocess
import sys
import tarfile
import tempfile
from contextlib import contextmanager
from pathlib import Path
from typing import Generator, Literal

import yaml

from mlflow.artifacts import download_artifacts
from mlflow.models.model import MLMODEL_FILE_NAME
from mlflow.utils.databricks_utils import DatabricksRuntimeVersion, get_databricks_runtime_version
from mlflow.utils.environment import _REQUIREMENTS_FILE_NAME
from mlflow.utils.logging_utils import eprint

EnvPackType = Literal["databricks_model_serving"]

_ARTIFACT_PATH = "_databricks"
_MODEL_VERSION_TAR = "model_version.tar"
_MODEL_ENVIRONMENT_TAR = "model_environment.tar"


def _tar(root_path: Path, tar_path: Path) -> tarfile.TarFile:
    """
    Package all files under root_path into a tar at tar_path, excluding __pycache__, *.pyc, and
    wheels_info.json.
    """

    def exclude(tarinfo: tarfile.TarInfo):
        name = tarinfo.name
        base = Path(name).name
        if "__pycache__" in name or base.endswith(".pyc") or base == "wheels_info.json":
            return None
        return tarinfo

    # Pull in symlinks
    with tarfile.open(tar_path, "w", dereference=True) as tar:
        tar.add(root_path, arcname=".", filter=exclude)
    return tar


# TODO: Check pip requirements using uv instead.
@contextmanager
def pack_env_for_databricks_model_serving(
    model_uri: str,
    *,
    enforce_pip_requirements: bool = False,
) -> Generator[str, None, None]:
    """
    Generate Databricks artifacts for fast deployment.

    Args:
        model_uri: The URI of the model to package.
        enforce_pip_requirements: Whether to enforce pip requirements installation.

    Yields:
        str: The path to the local artifacts directory containing the model artifacts and
            environment.

    Example:
        >>> with pack_env_for_databricks_model_serving("models:/my-model/1") as artifacts_dir:
        ...     # Use artifacts_dir here
        ...     pass
    """
    dbr_version = DatabricksRuntimeVersion.parse()
    if not dbr_version.is_client_image:
        raise ValueError(
            f"Serverless environment is required when packing environment for Databricks Model "
            f"Serving. Current version: {dbr_version}"
        )

    with tempfile.TemporaryDirectory() as temp_dir:
        # Download model artifacts. Keep this separate from temp_dir to avoid noise in packaged
        # artifacts.
        local_artifacts_dir = Path(download_artifacts(artifact_uri=model_uri))

        # Check runtime version consistency
        # We read the MLmodel file directly instead of using Model.to_dict() because to_dict() adds
        # the current runtime version via get_databricks_runtime_version(), which would prevent us
        # from detecting runtime version mismatches.
        mlmodel_path = local_artifacts_dir / MLMODEL_FILE_NAME
        with open(mlmodel_path) as f:
            model_dict = yaml.safe_load(f)
        if "databricks_runtime" not in model_dict:
            raise ValueError(
                "Model must have been created in a Databricks runtime environment. "
                "Missing 'databricks_runtime' field in MLmodel file."
            )

        current_runtime = DatabricksRuntimeVersion.parse()
        model_runtime = DatabricksRuntimeVersion.parse(model_dict["databricks_runtime"])
        if current_runtime.major != model_runtime.major:
            raise ValueError(
                f"Runtime version mismatch. Model was created with runtime "
                f"{model_dict['databricks_runtime']} (major version {model_runtime.major}), "
                f"but current runtime is {get_databricks_runtime_version()} "
                f"(major version {current_runtime.major})"
            )

        if enforce_pip_requirements:
            eprint("Installing model requirements...")
            try:
                subprocess.run(
                    [
                        sys.executable,
                        "-m",
                        "pip",
                        "install",
                        "-r",
                        str(local_artifacts_dir / _REQUIREMENTS_FILE_NAME),
                    ],
                    check=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                )
            except subprocess.CalledProcessError as e:
                eprint("Error installing requirements:")
                eprint(e.stdout)
                raise

        # Package model artifacts and env into temp_dir/_databricks
        temp_artifacts_dir = Path(temp_dir) / _ARTIFACT_PATH
        temp_artifacts_dir.mkdir(exist_ok=False)
        _tar(local_artifacts_dir, temp_artifacts_dir / _MODEL_VERSION_TAR)
        _tar(Path(sys.prefix), temp_artifacts_dir / _MODEL_ENVIRONMENT_TAR)
        shutil.move(str(temp_artifacts_dir), local_artifacts_dir)

        yield str(local_artifacts_dir)



"""add step to inputs table

Revision ID: 6953534de441
Revises: 400f98739977
Create Date: 2025-02-13 11:50:07.098121

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "6953534de441"
down_revision = "400f98739977"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("inputs", schema=None) as batch_op:
        batch_op.add_column(sa.Column("step", sa.BigInteger(), nullable=False, server_default="0"))


def downgrade():
    pass



import ast
import re
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

import sqlalchemy
import sqlparse

from mlflow.exceptions import MlflowException
from mlflow.store.tracking.dbmodels.models import SqlLoggedModel
from mlflow.utils.search_utils import _join_in_comparison_tokens


class EntityType(Enum):
    ATTRIBUTE = "attributes"
    METRIC = "metrics"
    PARAM = "params"
    TAG = "tags"

    @classmethod
    def from_str(cls, s: str) -> "EntityType":
        if s == "attributes":
            return cls.ATTRIBUTE
        if s == "metrics":
            return cls.METRIC
        if s == "params":
            return cls.PARAM
        if s == "tags":
            return cls.TAG

        raise MlflowException.invalid_parameter_value(
            f"Invalid entity type: {s!r}. Expected one of {[e.value for e in cls]}."
        )


@dataclass
class Entity:
    type: EntityType
    key: str

    IDENTIFIER_RE = re.compile(r"^([a-z]+)\.(.+)$")

    def __repr__(self) -> str:
        return f"{self.type.value}.{self.key}"

    @classmethod
    def from_str(cls, s: str) -> "Entity":
        if m := Entity.IDENTIFIER_RE.match(s):
            return cls(
                type=EntityType.from_str(m.group(1)),
                key=m.group(2).strip("`"),
            )
        return cls(type=EntityType.ATTRIBUTE, key=SqlLoggedModel.ALIASES.get(s, s).strip("`"))

    def is_numeric(self) -> bool:
        """
        Does this entity represent a numeric column?
        """
        return self.type == EntityType.METRIC or (
            self.type == EntityType.ATTRIBUTE and SqlLoggedModel.is_numeric(self.key)
        )

    def validate_op(self, op: str) -> None:
        numeric_ops = ("<", "<=", ">", ">=", "=", "!=")
        string_ops = ("=", "!=", "LIKE", "ILIKE", "IN", "NOT IN")
        ops = numeric_ops if self.is_numeric() else string_ops
        if op not in ops:
            raise MlflowException.invalid_parameter_value(
                f"Invalid comparison operator for {self}: {op!r}. Expected one of {string_ops}."
            )


@dataclass
class Comparison:
    entity: Entity
    op: str
    value: Union[str, float]


def parse_filter_string(filter_string: Optional[str]) -> list[Comparison]:
    if not filter_string:
        return []
    try:
        parsed = sqlparse.parse(filter_string)
    except Exception as e:
        raise MlflowException.invalid_parameter_value(
            f"Invalid filter string: {filter_string!r}. {e!r}"
        ) from e

    if len(parsed) != 1:
        raise MlflowException.invalid_parameter_value(
            f"Invalid filter string: {filter_string!r}. Expected a single SQL expression.",
        )

    comparisons: list[sqlalchemy.BinaryExpression] = []
    for stmt in _join_in_comparison_tokens(parsed[0].tokens):
        # while index < len(statements):
        if isinstance(stmt, sqlparse.sql.Comparison):
            non_whitespace_tokens = [str(t) for t in stmt.tokens if not t.is_whitespace]
            if len(non_whitespace_tokens) != 3:
                raise MlflowException.invalid_parameter_value(
                    f"Invalid comparison: {stmt}. Expected a comparison with 3 tokens."
                )
            identifier, op, value = non_whitespace_tokens
            entity = Entity.from_str(identifier)
            entity.validate_op(op)
            value = float(value) if entity.is_numeric() else value.strip("'")
            if entity.is_numeric():
                value = float(value)
            else:
                if value.startswith("(") and value.endswith(")"):
                    value = ast.literal_eval(value)
                    value = (value,) if isinstance(value, str) else value
                else:
                    value = value.strip("'")
            comparisons.append(Comparison(entity=entity, op=op, value=value))
        elif stmt.value.strip().upper() == "AND":
            # Do nothing, this is just a separator
            pass
        else:
            raise MlflowException.invalid_parameter_value(
                f"Invalid filter string: {filter_string!r}. Expected a list of comparisons "
                f"separated by 'AND' (e.g. 'metrics.loss > 0.1 AND params.lr = 0.01')."
            )

    return comparisons



import base64
import json
import logging
from typing import Any, Optional, Union

from mlflow.types.chat import (
    ChatMessage,
    ChatTool,
    Function,
    FunctionToolDefinition,
    ImageContentPart,
    ImageUrl,
    TextContentPart,
    ToolCall,
)

_logger = logging.getLogger(__name__)


def convert_message_to_mlflow_chat(message: dict[str, Any]) -> ChatMessage:
    """
    Convert Bedrock Converse API's message object into MLflow's standard format (OpenAI compatible).

    Ref: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Message.html

    Args:
        message: Bedrock Converse API's message object.

    Returns:
        ChatMessage: MLflow's standard chat message object.
    """
    role = message["role"]
    contents = []
    tool_calls = []
    tool_call_id = None
    for content in message["content"]:
        if tool_call := content.get("toolUse"):
            input = tool_call.get("input")
            tool_calls.append(
                ToolCall(
                    id=tool_call["toolUseId"],
                    function=Function(
                        name=tool_call["name"],
                        arguments=input if isinstance(input, str) else json.dumps(input),
                    ),
                    type="function",
                )
            )
        elif tool_result := content.get("toolResult"):
            tool_call_id = tool_result["toolUseId"]
            # "tool_result" content corresponds to the "tool" message in OpenAI.
            # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ToolResultContentBlock.html
            role = "tool"
            for content in tool_result["content"]:
                parsed_content = _parse_content(content)
                if parsed_content:
                    contents.append(parsed_content)
        else:
            parsed_content = _parse_content(content)
            if parsed_content:
                contents.append(parsed_content)

    message = ChatMessage(role=role, content=contents)
    if tool_calls:
        message.tool_calls = tool_calls
    if tool_call_id:
        message.tool_call_id = tool_call_id
    return message


def _parse_content(content: dict[str, Any]) -> Optional[Union[TextContentPart, ImageContentPart]]:
    """
    Parse a single content block in the Bedrock message object.

    Some content types like video and document are not supported by OpenAI's spec. This
    function returns None for those content types.

    Ref: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ContentBlock.html
    """
    if text := content.get("text"):
        return TextContentPart(text=text, type="text")
    elif json_content := content.get("json"):
        return TextContentPart(text=json.dumps(json_content), type="text")
    elif image := content.get("image"):
        # Bedrock support passing images in both raw bytes and base64 encoded strings.
        # OpenAI spec only supports base64 encoded images, so we encode the raw bytes to base64.
        # https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_ImageBlock.html
        image_bytes = image["source"]["bytes"]
        if isinstance(image_bytes, bytes):
            data = base64.b64encode(image_bytes).decode("utf-8")
        else:
            data = image_bytes
        format = "image/" + image["format"]
        image_url = ImageUrl(url=f"data:{format};base64,{data}", detail="auto")
        return ImageContentPart(type="image_url", image_url=image_url)
    # NB: Video and Document content type are not supported by OpenAI's spec, so recording as text.
    else:
        _logger.debug(f"Received an unsupported content type: {list(content.keys())[0]}")
        return None


def convert_tool_to_mlflow_chat_tool(tool: dict[str, Any]) -> ChatTool:
    """
    Convert Bedrock tool definition into MLflow's standard format (OpenAI compatible).

    Ref: https://docs.aws.amazon.com/bedrock/latest/APIReference/API_runtime_Tool.html

    Args:
        tool: A dictionary represents a single tool definition in the input request.

    Returns:
        ChatTool: MLflow's standard tool definition object.
    """
    tool_spec = tool["toolSpec"]
    return ChatTool(
        type="function",
        function=FunctionToolDefinition(
            name=tool_spec["name"],
            description=tool_spec.get("description"),
            parameters=tool_spec["inputSchema"].get("json"),
        ),
    )



from typing import Any, Optional

from mlflow.entities.span import LiveSpan, Span
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import INVALID_STATE
from mlflow.tracing.trace_manager import InMemoryTraceManager


def copy_trace_to_experiment(
    trace_dict: dict[str, Any], experiment_id: Optional[str] = None
) -> str:
    """
    Copy the given trace to the current experiment.
    The copied trace will have a new trace ID and location metadata.

    Args:
        trace_dict: The trace dictionary returned from model serving endpoint.
            This can be either V2 or V3 trace.
        experiment_id: The ID of the experiment to copy the trace to.
            If not provided, the trace will be copied to the current experiment.
    """
    new_trace_id = None
    new_root_span = None
    trace_manager = InMemoryTraceManager.get_instance()
    spans = [Span.from_dict(span_dict) for span_dict in trace_dict["data"]["spans"]]

    # Create a copy of spans in the current experiment
    for old_span in spans:
        new_span = LiveSpan.from_immutable_span(
            span=old_span,
            parent_span_id=old_span.parent_id,
            trace_id=new_trace_id,
            experiment_id=experiment_id,
            # Don't close the root span until the end so that we only export the trace
            # after all spans are copied.
            end_trace=old_span.parent_id is not None,
        )
        trace_manager.register_span(new_span)
        if old_span.parent_id is None:
            new_root_span = new_span
            new_trace_id = new_span.trace_id

    if new_trace_id is None:
        raise MlflowException(
            "Root span not found in the trace. Perhaps the trace data is corrupted.",
            error_code=INVALID_STATE,
        )

    user_tags = {k: v for k, v in trace_dict["info"]["tags"].items() if not k.startswith("mlflow.")}
    with trace_manager.get_trace(trace_id=new_trace_id) as trace:
        trace.info.tags.update(user_tags)

    # Close the root span triggers the trace export.
    new_root_span.end(end_time_ns=spans[0].end_time_ns)
    return new_trace_id



import functools
import importlib.metadata
import json
import logging
import warnings
from typing import Any, AsyncIterator, Iterator

from packaging.version import Version

import mlflow
from mlflow.entities import SpanType
from mlflow.entities.span import LiveSpan
from mlflow.entities.span_event import SpanEvent
from mlflow.entities.span_status import SpanStatusCode
from mlflow.exceptions import MlflowException
from mlflow.openai.constant import FLAVOR_NAME
from mlflow.openai.utils.chat_schema import set_span_chat_attributes
from mlflow.tracing.constant import (
    STREAM_CHUNK_EVENT_NAME_FORMAT,
    STREAM_CHUNK_EVENT_VALUE_KEY,
    SpanAttributeKey,
    TokenUsageKey,
    TraceMetadataKey,
)
from mlflow.tracing.fluent import start_span_no_context
from mlflow.tracing.trace_manager import InMemoryTraceManager
from mlflow.tracing.utils import TraceJSONEncoder
from mlflow.utils.annotations import experimental
from mlflow.utils.autologging_utils import autologging_integration
from mlflow.utils.autologging_utils.config import AutoLoggingConfig
from mlflow.utils.autologging_utils.safety import safe_patch

_logger = logging.getLogger(__name__)


@experimental(version="2.14.0")
def autolog(
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    log_traces=True,
):
    """
    Enables (or disables) and configures autologging from OpenAI to MLflow.
    Raises :py:class:`MlflowException <mlflow.exceptions.MlflowException>`
    if the OpenAI version < 1.0.

    Args:
        disable: If ``True``, disables the OpenAI autologging integration. If ``False``,
            enables the OpenAI autologging integration.
        exclusive: If ``True``, autologged content is not logged to user-created fluent runs.
            If ``False``, autologged content is logged to the active fluent run,
            which may be user-created.
        disable_for_unsupported_versions: If ``True``, disable autologging for versions of
            OpenAI that have not been tested against this version of the MLflow
            client or are incompatible.
        silent: If ``True``, suppress all event logs and warnings from MLflow during OpenAI
            autologging. If ``False``, show all events and warnings during OpenAI
            autologging.
        log_traces: If ``True``, traces are logged for OpenAI models. If ``False``, no traces are
            collected during inference. Default to ``True``.
    """
    if Version(importlib.metadata.version("openai")).major < 1:
        raise MlflowException("OpenAI autologging is only supported for openai >= 1.0.0")

    # This needs to be called before doing any safe-patching (otherwise safe-patch will be no-op).
    # TODO: since this implementation is inconsistent, explore a universal way to solve the issue.
    _autolog(
        disable=disable,
        exclusive=exclusive,
        disable_for_unsupported_versions=disable_for_unsupported_versions,
        silent=silent,
        log_traces=log_traces,
    )

    # Tracing OpenAI Agent SDK. This has to be done outside the function annotated with
    # `@autologging_integration` because the function is not executed when `disable=True`.
    try:
        from mlflow.openai._agent_tracer import (
            add_mlflow_trace_processor,
            remove_mlflow_trace_processor,
        )

        if log_traces and not disable:
            add_mlflow_trace_processor()
        else:
            remove_mlflow_trace_processor()
    except ImportError:
        pass


# This is required by mlflow.autolog()
autolog.integration_name = FLAVOR_NAME


# NB: The @autologging_integration annotation must be applied here, and the callback injection
# needs to happen outside the annotated function. This is because the annotated function is NOT
# executed when disable=True is passed. This prevents us from removing our callback and patching
# when autologging is turned off.
@autologging_integration(FLAVOR_NAME)
def _autolog(
    disable=False,
    exclusive=False,
    disable_for_unsupported_versions=False,
    silent=False,
    log_traces=True,
):
    from openai.resources.chat.completions import AsyncCompletions as AsyncChatCompletions
    from openai.resources.chat.completions import Completions as ChatCompletions
    from openai.resources.completions import AsyncCompletions, Completions
    from openai.resources.embeddings import AsyncEmbeddings, Embeddings

    for task in (ChatCompletions, Completions, Embeddings):
        safe_patch(FLAVOR_NAME, task, "create", patched_call)

    if hasattr(ChatCompletions, "parse"):
        # In openai>=1.92.0, `ChatCompletions` has a `parse` method:
        # https://github.com/openai/openai-python/commit/0e358ed66b317038705fb38958a449d284f3cb88
        safe_patch(FLAVOR_NAME, ChatCompletions, "parse", patched_call)

    for task in (AsyncChatCompletions, AsyncCompletions, AsyncEmbeddings):
        safe_patch(FLAVOR_NAME, task, "create", async_patched_call)

    if hasattr(AsyncChatCompletions, "parse"):
        # In openai>=1.92.0, `AsyncChatCompletions` has a `parse` method:
        # https://github.com/openai/openai-python/commit/0e358ed66b317038705fb38958a449d284f3cb88
        safe_patch(FLAVOR_NAME, AsyncChatCompletions, "parse", async_patched_call)

    try:
        from openai.resources.beta.chat.completions import AsyncCompletions, Completions
    except ImportError:
        pass
    else:
        safe_patch(FLAVOR_NAME, Completions, "parse", patched_call)
        safe_patch(FLAVOR_NAME, AsyncCompletions, "parse", async_patched_call)

    try:
        from openai.resources.responses import AsyncResponses, Responses
    except ImportError:
        pass
    else:
        safe_patch(FLAVOR_NAME, Responses, "create", patched_call)
        safe_patch(FLAVOR_NAME, AsyncResponses, "create", async_patched_call)
        safe_patch(FLAVOR_NAME, AsyncResponses, "parse", async_patched_call)
        safe_patch(FLAVOR_NAME, Responses, "parse", patched_call)

    # Patch Swarm agent to generate traces
    try:
        from swarm import Swarm

        warnings.warn(
            "Autologging for OpenAI Swarm is deprecated and will be removed in a future release. "
            "OpenAI Agent SDK is drop-in replacement for agent building and is supported by "
            "MLflow autologging. Please refer to the OpenAI Agent SDK documentation "
            "(https://github.com/openai/openai-agents-python) for more details.",
            category=FutureWarning,
            stacklevel=2,
        )

        safe_patch(
            FLAVOR_NAME,
            Swarm,
            "get_chat_completion",
            patched_agent_get_chat_completion,
        )

        safe_patch(
            FLAVOR_NAME,
            Swarm,
            "run",
            patched_swarm_run,
        )
    except ImportError:
        pass


def _get_span_type(task: type) -> str:
    from openai.resources.chat.completions import AsyncCompletions as AsyncChatCompletions
    from openai.resources.chat.completions import Completions as ChatCompletions
    from openai.resources.completions import AsyncCompletions, Completions
    from openai.resources.embeddings import AsyncEmbeddings, Embeddings

    span_type_mapping = {
        ChatCompletions: SpanType.CHAT_MODEL,
        AsyncChatCompletions: SpanType.CHAT_MODEL,
        Completions: SpanType.LLM,
        AsyncCompletions: SpanType.LLM,
        Embeddings: SpanType.EMBEDDING,
        AsyncEmbeddings: SpanType.EMBEDDING,
    }

    try:
        # Only available in openai>=1.40.0
        from openai.resources.beta.chat.completions import (
            AsyncCompletions as BetaAsyncChatCompletions,
        )
        from openai.resources.beta.chat.completions import Completions as BetaChatCompletions

        span_type_mapping[BetaChatCompletions] = SpanType.CHAT_MODEL
        span_type_mapping[BetaAsyncChatCompletions] = SpanType.CHAT_MODEL
    except ImportError:
        _logger.debug(
            "Failed to import `BetaChatCompletions` or `BetaAsyncChatCompletions`", exc_info=True
        )

    try:
        # Responses API only available in openai>=1.66.0
        from openai.resources.responses import AsyncResponses, Responses

        span_type_mapping[Responses] = SpanType.CHAT_MODEL
        span_type_mapping[AsyncResponses] = SpanType.CHAT_MODEL
    except ImportError:
        pass

    return span_type_mapping.get(task, SpanType.UNKNOWN)


def _try_parse_raw_response(response: Any) -> Any:
    """
    As documented at https://github.com/openai/openai-python/tree/52357cff50bee57ef442e94d78a0de38b4173fc2?tab=readme-ov-file#accessing-raw-response-data-eg-headers,
    a `LegacyAPIResponse` (https://github.com/openai/openai-python/blob/52357cff50bee57ef442e94d78a0de38b4173fc2/src/openai/_legacy_response.py#L45)
    object is returned when the `create` method is invoked with `with_raw_response`.
    """
    try:
        from openai._legacy_response import LegacyAPIResponse
    except ImportError:
        _logger.debug("Failed to import `LegacyAPIResponse` from `openai._legacy_response`")
        return response
    if isinstance(response, LegacyAPIResponse):
        try:
            # `parse` returns either a `pydantic.BaseModel` or a `openai.Stream` object
            # depending on whether the request has a `stream` parameter set to `True`.
            return response.parse()
        except Exception as e:
            _logger.debug(f"Failed to parse {response} (type: {response.__class__}): {e}")

    return response


def patched_call(original, self, *args, **kwargs):
    config = AutoLoggingConfig.init(flavor_name=mlflow.openai.FLAVOR_NAME)
    active_run = mlflow.active_run()
    run_id = active_run.info.run_id if active_run else None

    if config.log_traces:
        span = _start_span(self, kwargs, run_id)

    # Execute the original function
    try:
        raw_result = original(self, *args, **kwargs)
    except Exception as e:
        if config.log_traces:
            _end_span_on_exception(span, e)
        raise

    if config.log_traces:
        _end_span_on_success(span, kwargs, raw_result)

    return raw_result


async def async_patched_call(original, self, *args, **kwargs):
    config = AutoLoggingConfig.init(flavor_name=mlflow.openai.FLAVOR_NAME)
    active_run = mlflow.active_run()
    run_id = active_run.info.run_id if active_run else None

    if config.log_traces:
        span = _start_span(self, kwargs, run_id)

    # Execute the original function
    try:
        raw_result = await original(self, *args, **kwargs)
    except Exception as e:
        if config.log_traces:
            _end_span_on_exception(span, e)
        raise

    if config.log_traces:
        _end_span_on_success(span, kwargs, raw_result)

    return raw_result


def _start_span(
    instance: Any,
    inputs: dict[str, Any],
    run_id: str,
):
    # Record input parameters to attributes
    attributes = {k: v for k, v in inputs.items() if k not in ("messages", "input")}

    # If there is an active span, create a child span under it, otherwise create a new trace
    span = start_span_no_context(
        name=instance.__class__.__name__,
        span_type=_get_span_type(instance.__class__),
        inputs=inputs,
        attributes=attributes,
    )

    # Associate run ID to the trace manually, because if a new run is created by
    # autologging, it is not set as the active run thus not automatically
    # associated with the trace.
    if run_id is not None:
        tm = InMemoryTraceManager().get_instance()
        tm.set_trace_metadata(span.trace_id, TraceMetadataKey.SOURCE_RUN, run_id)

    return span


def _end_span_on_success(span: LiveSpan, inputs: dict[str, Any], raw_result: Any):
    from openai import AsyncStream, Stream

    result = _try_parse_raw_response(raw_result)

    if isinstance(result, Stream):
        # If the output is a stream, we add a hook to store the intermediate chunks
        # and then log the outputs as a single artifact when the stream ends
        def _stream_output_logging_hook(stream: Iterator) -> Iterator:
            output = []
            for i, chunk in enumerate(stream):
                _add_span_event(span, i, chunk)
                output.append(chunk)
                yield chunk
            _process_last_chunk(span, chunk, inputs, output)

        result._iterator = _stream_output_logging_hook(result._iterator)
    elif isinstance(result, AsyncStream):

        async def _stream_output_logging_hook(stream: AsyncIterator) -> AsyncIterator:
            output = []
            async for chunk in stream:
                _add_span_event(span, len(output), chunk)
                output.append(chunk)
                yield chunk
            _process_last_chunk(span, chunk, inputs, output)

        result._iterator = _stream_output_logging_hook(result._iterator)
    else:
        try:
            set_span_chat_attributes(span, inputs, result)
            span.end(outputs=result)
        except Exception as e:
            _logger.warning(f"Encountered unexpected error when ending trace: {e}", exc_info=True)


def _process_last_chunk(span: LiveSpan, chunk: Any, inputs: dict[str, Any], output: list[Any]):
    if _is_responses_final_event(chunk):
        output = chunk.response
    else:
        # Reconstruct a completion object from streaming chunks
        output = _reconstruct_completion_from_stream(output)

        # Set usage information on span if available
        if usage := getattr(chunk, "usage", None):
            usage_dict = {
                TokenUsageKey.INPUT_TOKENS: usage.prompt_tokens,
                TokenUsageKey.OUTPUT_TOKENS: usage.completion_tokens,
                TokenUsageKey.TOTAL_TOKENS: usage.total_tokens,
            }
            span.set_attribute(SpanAttributeKey.CHAT_USAGE, usage_dict)

    _end_span_on_success(span, inputs, output)


def _reconstruct_completion_from_stream(chunks: list[Any]) -> Any:
    """
    Reconstruct a completion object from streaming chunks.

    This preserves the structure and metadata that would be present in a non-streaming
    completion response, including ID, model, timestamps, usage, etc.
    """
    if not chunks:
        return None

    if chunks[0].object == "text_completion":
        # Handling for the deprecated Completions API. Keep the legacy behavior for now.
        def _extract_content(chunk: Any) -> str:
            if not chunk.choices:
                return ""
            return chunk.choices[0].text or ""

        return "".join(map(_extract_content, chunks))

    if chunks[0].object != "chat.completion.chunk":
        return chunks  # Ignore non-chat chunks

    from openai.types.chat import ChatCompletion
    from openai.types.chat.chat_completion import Choice
    from openai.types.chat.chat_completion_message import ChatCompletionMessage

    # Build the base message
    def _extract_content(chunk: Any) -> str:
        if not chunk.choices:
            return ""
        return chunk.choices[0].delta.content or ""

    message = ChatCompletionMessage(
        role="assistant", content="".join(map(_extract_content, chunks))
    )

    # Extract metadata from the last chunk
    last_chunk = chunks[-1]
    finish_reason = "stop"
    if choices := getattr(last_chunk, "choices", None):
        if chunk_choice := choices[0]:
            finish_reason = getattr(chunk_choice, "finish_reason") or finish_reason

    choice = Choice(index=0, message=message, finish_reason=finish_reason)

    # Build the completion dict
    return ChatCompletion(
        id=last_chunk.id,
        choices=[choice],
        created=last_chunk.created,
        model=last_chunk.model,
        object="chat.completion",
        system_fingerprint=last_chunk.system_fingerprint,
        usage=last_chunk.usage,
    )


def _is_responses_final_event(chunk: Any) -> bool:
    try:
        from openai.types.responses import ResponseCompletedEvent

        return isinstance(chunk, ResponseCompletedEvent)
    except ImportError:
        return False


def _end_span_on_exception(span: LiveSpan, e: Exception):
    try:
        span.add_event(SpanEvent.from_exception(e))
        span.end(status=SpanStatusCode.ERROR)
    except Exception as inner_e:
        _logger.warning(f"Encountered unexpected error when ending trace: {inner_e}")


def _add_span_event(span: LiveSpan, index: int, chunk: Any):
    span.add_event(
        SpanEvent(
            name=STREAM_CHUNK_EVENT_NAME_FORMAT.format(index=index),
            # OpenTelemetry SpanEvent only support str-str key-value pairs for attributes
            attributes={STREAM_CHUNK_EVENT_VALUE_KEY: json.dumps(chunk, cls=TraceJSONEncoder)},
        )
    )


def patched_agent_get_chat_completion(original, self, *args, **kwargs):
    """
    Patch the `get_chat_completion` method of the ChatCompletion object.
    OpenAI autolog already handles the raw completion request, but tracing
    the swarm's method is useful to track other parameters like agent name.
    """
    agent = kwargs.get("agent") or args[0]

    # Patch agent's functions to generate traces. Function calls only happen
    # after the first completion is generated because of the design of
    # function calling. Therefore, we can safely patch the tool functions here
    # within get_chat_completion() hook.
    # We cannot patch functions during the agent's initialization because the
    # agent's functions can be modified after the agent is created.
    def function_wrapper(fn):
        if "context_variables" in fn.__code__.co_varnames:

            def wrapper(*args, **kwargs):
                # NB: Swarm uses `func.__code__.co_varnames` to inspect if the provided
                # tool function includes 'context_variables' parameter in the signature
                # and ingest the global context variables if so. Wrapping the function
                # with mlflow.trace() will break this.
                # The co_varnames is determined based on the local variables of the
                # function, so we workaround this by declaring it here as a local variable.
                context_variables = kwargs.get("context_variables", {})  # noqa: F841
                return mlflow.trace(
                    fn,
                    name=f"{agent.name}.{fn.__name__}",
                    span_type=SpanType.TOOL,
                )(*args, **kwargs)
        else:

            def wrapper(*args, **kwargs):
                return mlflow.trace(
                    fn,
                    name=f"{agent.name}.{fn.__name__}",
                    span_type=SpanType.TOOL,
                )(*args, **kwargs)

        wrapped = functools.wraps(fn)(wrapper)
        wrapped._is_mlflow_traced = True  # Marker to avoid double tracing
        return wrapped

    agent.functions = [
        function_wrapper(fn) if not hasattr(fn, "_is_mlflow_traced") else fn
        for fn in agent.functions
    ]

    traced_fn = mlflow.trace(
        original, name=f"{agent.name}.get_chat_completion", span_type=SpanType.CHAIN
    )
    return traced_fn(self, *args, **kwargs)


def patched_swarm_run(original, self, *args, **kwargs):
    """
    Patched version of `run` method of the Swarm object.
    """
    traced_fn = mlflow.trace(original, span_type=SpanType.AGENT)
    return traced_fn(self, *args, **kwargs)



"""
Base Plugin Classes for Genesis-Flow Plugin System

Defines the interface and base classes that all plugins must implement.
"""

import logging
from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Any, Callable
from dataclasses import dataclass
from enum import Enum

logger = logging.getLogger(__name__)

class PluginType(Enum):
    """Types of plugins supported by Genesis-Flow."""
    FRAMEWORK = "framework"      # ML framework integration (PyTorch, TensorFlow, etc.)
    LOGGING = "logging"         # Custom logging backends
    MODEL_REGISTRY = "registry" # Model registry integrations
    DEPLOYMENT = "deployment"   # Deployment platforms
    ARTIFACT = "artifact"       # Artifact storage backends
    UI = "ui"                   # UI extensions
    CUSTOM = "custom"           # Custom extensions

@dataclass
class PluginMetadata:
    """Metadata for a plugin."""
    name: str
    version: str
    description: str
    author: str
    plugin_type: PluginType
    dependencies: List[str]
    optional_dependencies: List[str]
    min_genesis_flow_version: str
    max_genesis_flow_version: Optional[str] = None
    homepage: Optional[str] = None
    documentation: Optional[str] = None
    license: Optional[str] = None
    tags: List[str] = None

    def __post_init__(self):
        if self.tags is None:
            self.tags = []

class PluginState(Enum):
    """States a plugin can be in."""
    DISCOVERED = "discovered"
    LOADED = "loaded"
    ENABLED = "enabled" 
    DISABLED = "disabled"
    FAILED = "failed"

class BasePlugin(ABC):
    """
    Base class for all Genesis-Flow plugins.
    
    All plugins must inherit from this class and implement the required methods.
    """
    
    def __init__(self, metadata: PluginMetadata):
        self.metadata = metadata
        self.state = PluginState.DISCOVERED
        self._logger = logging.getLogger(f"mlflow.plugins.{metadata.name}")
        self._hooks = {}
        
    @abstractmethod
    def load(self) -> bool:
        """
        Load the plugin and perform any necessary initialization.
        
        Returns:
            True if loading was successful, False otherwise
        """
        pass
    
    @abstractmethod
    def enable(self) -> bool:
        """
        Enable the plugin for use.
        
        Returns:
            True if enabling was successful, False otherwise
        """
        pass
    
    @abstractmethod
    def disable(self) -> bool:
        """
        Disable the plugin.
        
        Returns:
            True if disabling was successful, False otherwise
        """
        pass
    
    @abstractmethod
    def unload(self) -> bool:
        """
        Unload the plugin and clean up resources.
        
        Returns:
            True if unloading was successful, False otherwise
        """
        pass
    
    def is_compatible(self, genesis_flow_version: str) -> bool:
        """
        Check if the plugin is compatible with the current Genesis-Flow version.
        
        Args:
            genesis_flow_version: Current Genesis-Flow version
            
        Returns:
            True if compatible, False otherwise
        """
        from packaging.version import Version
        
        try:
            current = Version(genesis_flow_version)
            min_version = Version(self.metadata.min_genesis_flow_version)
            
            if current < min_version:
                return False
                
            if self.metadata.max_genesis_flow_version:
                max_version = Version(self.metadata.max_genesis_flow_version)
                if current > max_version:
                    return False
                    
            return True
            
        except Exception as e:
            self._logger.warning(f"Version compatibility check failed: {e}")
            return False
    
    def check_dependencies(self) -> bool:
        """
        Check if all required dependencies are available.
        
        Returns:
            True if all dependencies are satisfied, False otherwise
        """
        try:
            import importlib
            
            for dep in self.metadata.dependencies:
                try:
                    importlib.import_module(dep)
                except ImportError:
                    self._logger.error(f"Required dependency '{dep}' not found")
                    return False
            
            # Check optional dependencies and log warnings
            for dep in self.metadata.optional_dependencies:
                try:
                    importlib.import_module(dep)
                except ImportError:
                    self._logger.warning(f"Optional dependency '{dep}' not found")
            
            return True
            
        except Exception as e:
            self._logger.error(f"Dependency check failed: {e}")
            return False
    
    def register_hook(self, event: str, callback: Callable):
        """Register a callback for a specific event."""
        if event not in self._hooks:
            self._hooks[event] = []
        self._hooks[event].append(callback)
    
    def trigger_hook(self, event: str, *args, **kwargs):
        """Trigger all callbacks for a specific event."""
        if event in self._hooks:
            for callback in self._hooks[event]:
                try:
                    callback(*args, **kwargs)
                except Exception as e:
                    self._logger.error(f"Hook callback failed for {event}: {e}")
    
    def get_info(self) -> Dict[str, Any]:
        """Get plugin information."""
        return {
            "name": self.metadata.name,
            "version": self.metadata.version,
            "description": self.metadata.description,
            "author": self.metadata.author,
            "type": self.metadata.plugin_type.value,
            "state": self.state.value,
            "dependencies": self.metadata.dependencies,
            "optional_dependencies": self.metadata.optional_dependencies,
            "tags": self.metadata.tags,
        }

class FrameworkPlugin(BasePlugin):
    """
    Base class for ML framework plugins.
    
    Framework plugins provide integration with ML libraries like PyTorch, TensorFlow, etc.
    """
    
    def __init__(self, metadata: PluginMetadata):
        super().__init__(metadata)
        self._module_path = None
        self._lazy_loader = None
    
    @abstractmethod
    def get_module_path(self) -> str:
        """
        Get the module path for the framework integration.
        
        Returns:
            Module path (e.g., "mlflow.pytorch")
        """
        pass
    
    @abstractmethod
    def get_autolog_functions(self) -> Dict[str, Callable]:
        """
        Get autologging functions provided by this framework.
        
        Returns:
            Dict mapping function names to callables
        """
        pass
    
    @abstractmethod
    def get_save_functions(self) -> Dict[str, Callable]:
        """
        Get model saving functions provided by this framework.
        
        Returns:
            Dict mapping function names to callables
        """
        pass
    
    @abstractmethod
    def get_load_functions(self) -> Dict[str, Callable]:
        """
        Get model loading functions provided by this framework.
        
        Returns:
            Dict mapping function names to callables
        """
        pass
    
    def load(self) -> bool:
        """Load the framework plugin."""
        try:
            if not self.check_dependencies():
                self.state = PluginState.FAILED
                return False
            
            self._module_path = self.get_module_path()
            
            # Create lazy loader for the framework
            from mlflow.utils.lazy_load import LazyLoader
            import mlflow
            
            self._lazy_loader = LazyLoader(
                self._module_path,
                mlflow.__dict__,
                self._module_path
            )
            
            self.state = PluginState.LOADED
            self._logger.info(f"Framework plugin {self.metadata.name} loaded")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to load framework plugin {self.metadata.name}: {e}")
            self.state = PluginState.FAILED
            return False
    
    def enable(self) -> bool:
        """Enable the framework plugin."""
        try:
            if self.state != PluginState.LOADED:
                if not self.load():
                    return False
            
            # Register the lazy loader in mlflow namespace
            import mlflow
            setattr(mlflow, self.metadata.name, self._lazy_loader)
            
            self.state = PluginState.ENABLED
            self._logger.info(f"Framework plugin {self.metadata.name} enabled")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to enable framework plugin {self.metadata.name}: {e}")
            self.state = PluginState.FAILED
            return False
    
    def disable(self) -> bool:
        """Disable the framework plugin."""
        try:
            import mlflow
            
            if hasattr(mlflow, self.metadata.name):
                delattr(mlflow, self.metadata.name)
            
            self.state = PluginState.LOADED
            self._logger.info(f"Framework plugin {self.metadata.name} disabled")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to disable framework plugin {self.metadata.name}: {e}")
            return False
    
    def unload(self) -> bool:
        """Unload the framework plugin."""
        try:
            self.disable()
            
            self._lazy_loader = None
            self._module_path = None
            
            self.state = PluginState.DISCOVERED
            self._logger.info(f"Framework plugin {self.metadata.name} unloaded")
            return True
            
        except Exception as e:
            self._logger.error(f"Failed to unload framework plugin {self.metadata.name}: {e}")
            return False

class LoggingPlugin(BasePlugin):
    """Base class for custom logging backend plugins."""
    
    @abstractmethod
    def get_logging_backend(self):
        """Get the logging backend implementation."""
        pass

class ModelRegistryPlugin(BasePlugin):
    """Base class for model registry integration plugins."""
    
    @abstractmethod
    def get_registry_backend(self):
        """Get the model registry backend implementation."""
        pass

class DeploymentPlugin(BasePlugin):
    """Base class for deployment platform plugins."""
    
    @abstractmethod
    def get_deployment_backend(self):
        """Get the deployment backend implementation."""
        pass

class ArtifactPlugin(BasePlugin):
    """Base class for artifact storage plugins."""
    
    @abstractmethod
    def get_artifact_repository(self):
        """Get the artifact repository implementation."""
        pass


import logging
from dataclasses import dataclass
from typing import Any, Callable, Optional

import numpy as np

from mlflow.metrics.base import MetricValue
from mlflow.models.evaluation.base import EvaluationMetric

_logger = logging.getLogger(__name__)


@dataclass
class MetricDefinition:
    """
    A namedtuple representing a metric function and its properties.

    function : the metric function
    name : the name of the metric function
    index : the index of the function in the ``extra_metrics`` argument of mlflow.evaluate
    """

    function: Callable[..., Any]
    name: str
    index: int
    version: Optional[str] = None
    genai_metric_args: Optional[dict[str, Any]] = None

    @classmethod
    def from_index_and_metric(cls, index: int, metric: EvaluationMetric):
        return cls(
            function=metric.eval_fn,
            index=index,
            name=metric.name,
            version=metric.version,
            genai_metric_args=metric.genai_metric_args,
        )

    def evaluate(self, eval_fn_args) -> Optional[MetricValue]:
        """
        This function calls the metric function and performs validations on the returned
        result to ensure that they are in the expected format. It will warn and will not log metrics
        that are in the wrong format.

        Args:
            eval_fn_args: A dictionary of args needed to compute the eval metrics.

        Returns:
            MetricValue
        """
        if self.index < 0:
            exception_header = f"Did not log builtin metric '{self.name}' because it"
        else:
            exception_header = (
                f"Did not log metric '{self.name}' at index "
                f"{self.index} in the `extra_metrics` parameter because it"
            )

        metric: MetricValue = self.function(*eval_fn_args)

        def _is_numeric(value):
            return isinstance(value, (int, float, np.number))

        def _is_string(value):
            return isinstance(value, str)

        if metric is None:
            _logger.warning(f"{exception_header} returned None.")
            return

        if _is_numeric(metric):
            return MetricValue(aggregate_results={self.name: metric})

        if not isinstance(metric, MetricValue):
            _logger.warning(f"{exception_header} did not return a MetricValue.")
            return

        scores = metric.scores
        justifications = metric.justifications
        aggregates = metric.aggregate_results

        if scores is not None:
            if not isinstance(scores, list):
                _logger.warning(
                    f"{exception_header} must return MetricValue with scores as a list."
                )
                return
            if any(not (_is_numeric(s) or _is_string(s) or s is None) for s in scores):
                _logger.warning(
                    f"{exception_header} must return MetricValue with numeric or string scores."
                )
                return

        if justifications is not None:
            if not isinstance(justifications, list):
                _logger.warning(
                    f"{exception_header} must return MetricValue with justifications as a list."
                )
                return
            if any(not (_is_string(just) or just is None) for just in justifications):
                _logger.warning(
                    f"{exception_header} must return MetricValue with string justifications."
                )
                return

        if aggregates is not None:
            if not isinstance(aggregates, dict):
                _logger.warning(
                    f"{exception_header} must return MetricValue with aggregate_results as a dict."
                )
                return

            if any(
                not (isinstance(k, str) and (_is_numeric(v) or v is None))
                for k, v in aggregates.items()
            ):
                _logger.warning(
                    f"{exception_header} must return MetricValue with aggregate_results with "
                    "str keys and numeric values."
                )
                return

        return metric



from copy import deepcopy
from dataclasses import asdict, dataclass, field
from typing import Any, Optional


@dataclass
class Document:
    """
    An entity used in MLflow Tracing to represent retrieved documents in a RETRIEVER span.

    Args:
        page_content: The content of the document.
        metadata: A dictionary of metadata associated with the document.
        id: The ID of the document.
    """

    page_content: str
    metadata: dict[str, Any] = field(default_factory=dict)
    id: Optional[str] = None

    @classmethod
    def from_langchain_document(cls, document):
        # older versions of langchain do not have the id attribute
        id = getattr(document, "id", None)

        return cls(
            page_content=document.page_content,
            metadata=deepcopy(document.metadata),
            id=id,
        )

    @classmethod
    def from_llama_index_node_with_score(cls, node_with_score):
        metadata = {
            "score": node_with_score.get_score(),
            # update after setting score so that it can be
            # overridden if the user wishes to do so
            **deepcopy(node_with_score.metadata),
        }

        return cls(
            page_content=node_with_score.get_content(),
            metadata=metadata,
            id=node_with_score.node_id,
        )

    def to_dict(self):
        return asdict(self)
